/***************************************************************************
                          cron_export.c  -  description
                             -------------------
    begin                : Sat Oct 28 2000
    copyright            : (C) 2000 by Dominique Laigle
    email                : lainal@gmx.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef CRON_IMPEX_H
#include <cron_impex.h>
#endif

#ifndef _CRON_FLAT_LINE_T
#define _CRON_FLAT_LINE_T
typedef struct {
                 guint       month;
                 guint       day_month;
                 guchar      hhmi_id;
                 guchar      group_id;
                 guchar      action;
               } cron_flat_line_t;
#endif

#ifndef GRP_NONE
#define GRP_NONE 0xff
#endif

/* Donnes et variables statiques (porte de ce fichier uniquement, non exportes) */

static gchar str_cron_line[256];
static cron_flat_line_t *line_cron_struct,   /* Crontab full expanded */
                         buffer_line_1,
                         buffer_line_2;
static guint nb_lines_cron_struct, nb_cron_struct_allocated,
             the_month, the_day_month;

static guchar month, day_month, grp_action, the_hh_id,
              group_definition_mask[MAX_GROUPES];

static gint idx_cycle, idx_group, idx_date,idx_line, idx_same, idx_shift,
            retour_fonctions, file_cron_desc;

static day_schedule_t *the_schedule;

static const gchar * cron_en_tete = "\n\
###################################################\n\
#                                                 #\n\
#      File automatically built by GHeather       #\n\
#                                                 #\n\
###################################################\n\
SHELL=/bin/sh\n\
MAILTO=\"\"\n\n";

static const horaire_t table_horaires[97]={
{ 0, 0 }, { 0, 15 }, { 0, 30 }, { 0, 45 }, { 1, 0 }, { 1, 15 }, { 1, 30 }, { 1, 45 },
{ 2, 0 }, { 2, 15 }, { 2, 30 }, { 2, 45 }, { 3, 0 }, { 3, 15 }, { 3, 30 }, { 3, 45 },
{ 4, 0 }, { 4, 15 }, { 4, 30 }, { 4, 45 }, { 5, 0 }, { 5, 15 }, { 5, 30 }, { 5, 45 },
{ 6, 0 }, { 6, 15 }, { 6, 30 }, { 6, 45 }, { 7, 0 }, { 7, 15 }, { 7, 30 }, { 7, 45 },
{ 8, 0 }, { 8, 15 }, { 8, 30 }, { 8, 45 }, { 9, 0 }, { 9, 15 }, { 9, 30 }, { 9, 45 },
{ 10, 0 }, { 10, 15 }, { 10, 30 }, { 10, 45 }, { 11, 0 }, { 11, 15 }, { 11, 30 }, { 11, 45 },
{ 12, 0 }, { 12, 15 }, { 12, 30 }, { 12, 45 }, { 13, 0 }, { 13, 15 }, { 13, 30 }, { 13, 45 },
{ 14, 0 }, { 14, 15 }, { 14, 30 }, { 14, 45 }, { 15, 0 }, { 15, 15 }, { 15, 30 }, { 15, 45 },
{ 16, 0 }, { 16, 15 }, { 16, 30 }, { 16, 45 }, { 17, 0 }, { 17, 15 }, { 17, 30 }, { 17, 45 },
{ 18, 0 }, { 18, 15 }, { 18, 30 }, { 18, 45 }, { 19, 0 }, { 19, 15 }, { 19, 30 }, { 19, 45 },
{ 20, 0 }, { 20, 15 }, { 20, 30 }, { 20, 45 }, { 21, 0 }, { 21, 15 }, { 21, 30 }, { 21, 45 },
{ 22, 0 }, { 22, 15 }, { 22, 30 }, { 22, 45 }, { 23, 0 }, { 23, 15 }, { 23, 30 }, { 23, 45 },
{ 23, 59 } /* should be 0:0 the day after... hence this small trick ! */
};

/*----------------------------------------------------------------------------------*/

/* DEBUG PURPOSE ONLY */
/*
static void cron_impex_print_cron()
{
  gint idx;
  printf("\n%d Lines\n",nb_lines_cron_struct);
  for ( idx = 0; idx < nb_lines_cron_struct; idx++ )
    printf("hh:%d mm:%d - M:%u D:%u G:%d A:%d\n",
        table_horaires[line_cron_struct[idx].hhmi_id].heures,
				table_horaires[line_cron_struct[idx].hhmi_id].minutes,
				line_cron_struct[idx].month,
				line_cron_struct[idx].day_month,
				line_cron_struct[idx].group_id,
				line_cron_struct[idx].action);
}
*/

/*----------------------------------------------------------------------------------*/

static void cron_impex_expand_cron_struct( void )
{
  gint           nb_cycles, day_idx_begin, idx_insert, idx_shift;
  day_schedule_t the_date;
  gboolean       inserted;

  for ( idx_date = 0; idx_date < 366; idx_date++ )
    {
      the_date = the_schedule[idx_date];

      month = g_date_month( &the_date.date );
      day_month = g_date_day( &the_date.date );

      /* Record of the cycles. It is not that quick since we have to keep
         an afterwards time sort within the day for the others phases.
         Events sorting is mandatory for the cron scheduling atomicity.
       */

      day_idx_begin = nb_lines_cron_struct;

      for ( idx_group = 0; idx_group < MAX_GROUPES; idx_group++ )
        {
          nb_cycles = the_date.day_program.nb_cycles[idx_group];

          if ( nb_cycles != 0 ) /* day is candidate to expansion */
           {
             /* We will add 2 lines ( cycle start-stop ). It can be that
                before we must extend the allocated memory to the data
                structure. If so then extend straight of 1024 more rows.
             */

             if ( nb_lines_cron_struct >= nb_cron_struct_allocated - 16 )
              {
                line_cron_struct = g_renew( cron_flat_line_t,
                                          line_cron_struct,
                                          nb_cron_struct_allocated + 1024 );
                nb_cron_struct_allocated += 1024;
              }

             for ( idx_cycle = 0; idx_cycle < nb_cycles; idx_cycle++ )
             {
               buffer_line_1.hhmi_id =
                the_date.day_program.day_cycles[idx_group][idx_cycle].heure_debut.heures * 4 +
                the_date.day_program.day_cycles[idx_group][idx_cycle].heure_debut.minutes / 15;

               buffer_line_1.group_id = idx_group;
               buffer_line_1.action = CRON_IMPEX_ON;

               buffer_line_1.month = month;
               buffer_line_1.day_month = day_month;

               /* Where to insert the start cycle event ? */

               inserted = FALSE;
               line_cron_struct[nb_lines_cron_struct].hhmi_id = 0;

               for ( idx_insert = day_idx_begin;
                     ( idx_insert < nb_lines_cron_struct ) &&
                     ( inserted == FALSE );
                     idx_insert++ )
                  if ( buffer_line_1.hhmi_id < line_cron_struct[idx_insert].hhmi_id )
                   {
                     inserted = TRUE;

                     for ( idx_shift = nb_lines_cron_struct;
                           idx_shift > idx_insert; idx_shift-- )
                         line_cron_struct[idx_shift] =
                                            line_cron_struct[idx_shift - 1];
                     line_cron_struct[idx_insert] = buffer_line_1;
                   }

               if ( inserted == FALSE ) /* Then append to the end */
                 line_cron_struct[nb_lines_cron_struct] = buffer_line_1;

               /* Record for the stop of the cycle */

               buffer_line_2.hhmi_id =
                the_date.day_program.day_cycles[idx_group][idx_cycle].heure_fin.heures * 4 +
                the_date.day_program.day_cycles[idx_group][idx_cycle].heure_fin.minutes / 15;

               buffer_line_2.group_id = idx_group;
               buffer_line_2.action = CRON_IMPEX_OFF;

               buffer_line_2.month = month;
               buffer_line_2.day_month = day_month;

               /* Where to insert the stop cycle event ? */

               nb_lines_cron_struct++;
               line_cron_struct[nb_lines_cron_struct].hhmi_id = 0;

               inserted = FALSE;
               for ( idx_insert = day_idx_begin;
                     ( idx_insert < nb_lines_cron_struct ) &&
                     ( inserted == FALSE );
                     idx_insert++ )
                  if ( buffer_line_2.hhmi_id < line_cron_struct[idx_insert].hhmi_id )
                   {
                     inserted = TRUE;

                     for ( idx_shift = nb_lines_cron_struct;
                           idx_shift > idx_insert; idx_shift-- )
                         line_cron_struct[idx_shift] =
                                            line_cron_struct[idx_shift - 1];
                     line_cron_struct[idx_insert] = buffer_line_2;
                   }
               if ( inserted == FALSE ) /* Then append to the end */
                 line_cron_struct[nb_lines_cron_struct] = buffer_line_2;

               nb_lines_cron_struct++;
             }
           }
        }
    }
}

/*----------------------------------------------------------------------------------*/

static void cron_impex_merge_group_actions()
{
  gchar  env_valeur, *gp_var_env_nom, *gp_var_env_val,
         *gp_id = g_new0( gchar, 3 );
  guchar all_groups;
  gint i;
  gboolean env_ok;

  if ( nb_lines_cron_struct < 1 ) /* Nothing to do : goodby */
     return;

  /* Group definitions from the environment variables. If the environment
     is not set, then an error is raised and the application exits. */

  for ( idx_group = 0; idx_group < MAX_GROUPES; idx_group++ )
    {
       g_snprintf( gp_id, 2, "%d", idx_group );
       gp_var_env_nom = g_strconcat( "HEATHER_GP_", gp_id, NULL );

       gp_var_env_val = getenv( gp_var_env_nom );

       g_return_if_fail( gp_var_env_val != NULL );

       env_valeur = ( guchar ) strtol( gp_var_env_val, NULL, 0 );
       group_definition_mask[idx_group] = env_valeur;

       g_free( gp_var_env_nom );
    }

  g_free( gp_id );

  /* Test the coherence of the environment : all the bit assignments
     must be unique within the 4 groups. */

  all_groups = group_definition_mask[0];
  env_ok = TRUE;

  for ( i = 1; i < MAX_GROUPES; i++ )
    if ( ( all_groups & group_definition_mask[i] ) !=0 )
     {
       env_ok = FALSE;
       break;
     }
    else all_groups |= group_definition_mask[i];

  g_return_if_fail( env_ok == TRUE ); /* cycle definition is broken */


  /*
     This first scan performs the full group action encoding with the
     tracking of the interface status.
     The time stamps are not unique yet.
   */

  grp_action = 0;

  for ( idx_line = 0; idx_line < nb_lines_cron_struct; idx_line++ )
   {
     idx_group = line_cron_struct[idx_line].group_id;

     if ( line_cron_struct[idx_line].action == CRON_IMPEX_ON )
         grp_action |= group_definition_mask[idx_group];
     else
         grp_action &= ~group_definition_mask[idx_group];

     line_cron_struct[idx_line].action = grp_action;
     line_cron_struct[idx_line].group_id = GRP_NONE;
   }
}

/*----------------------------------------------------------------------------------*/

static void cron_impex_merge_time_stamp()
{
  gint   delta;

  for ( idx_line = 0; idx_line < nb_lines_cron_struct; idx_line++ )
   {
     /* Retrieving the "key" hhmm_id, month, day_month */

     the_hh_id = line_cron_struct[idx_line].hhmi_id;
     the_month = line_cron_struct[idx_line].month;
     the_day_month = line_cron_struct[idx_line].day_month;

     /* scan for the same key ie same time-stamp keeping the newest that
        always take in consideration the most groups. As the data are time
        sorted, the line to keep is the first encountered from the bottom
     */

     for ( idx_same = nb_lines_cron_struct - 1;
           ( idx_same > idx_line ) &&
           ( ( the_hh_id != line_cron_struct[idx_same].hhmi_id ) ||
             ( the_month != line_cron_struct[idx_same].month ) ||
             ( the_day_month != line_cron_struct[idx_same].day_month )
           );
           idx_same--
         );

     if ( idx_same > idx_line )
      {
        /* Data sorted : between idx_line and idx_same we only have the
          same time stamp. We just need to shift the struct to the top.*/

        delta = idx_same - idx_line;

        for ( idx_shift = idx_same; idx_shift < nb_lines_cron_struct; idx_shift++ )
              line_cron_struct[idx_shift-delta] = line_cron_struct[idx_shift];

         nb_lines_cron_struct -= delta;
      }
   }
}

/*----------------------------------------------------------------------------------*/

static void cron_impex_merge_days_within_month()
{
  for ( idx_line = 0; idx_line < nb_lines_cron_struct; idx_line++ )
   {
     /* Retrieving the "key" hhmm_id, month, day_month */

     the_hh_id = line_cron_struct[idx_line].hhmi_id;
     the_month = line_cron_struct[idx_line].month;
     grp_action = line_cron_struct[idx_line].action;

     /* Encode the first day within the month. day_month represents
     now the days as a bit-array 0x31 ... 1 */

     the_day_month = 0x1 << line_cron_struct[idx_line].day_month;

     for ( idx_same = idx_line + 1;
           idx_same < nb_lines_cron_struct; idx_same++ )
       {
         if ( ( the_hh_id == line_cron_struct[idx_same].hhmi_id ) &&
              ( the_month == line_cron_struct[idx_same].month ) &&
              ( grp_action == line_cron_struct[idx_same].action ) )
          {
            /* Combine the day with the first line encountered */

            the_day_month |= 0x1 << line_cron_struct[idx_same].day_month;

            /* The line can be removed with the shift */

            for ( idx_shift = idx_same;
                  idx_shift < nb_lines_cron_struct - 1; idx_shift++ )

            line_cron_struct[idx_shift] = line_cron_struct[idx_shift + 1];

            nb_lines_cron_struct -= 1;
            idx_same--;
          }
       }
     line_cron_struct[idx_line].day_month = the_day_month;
   }
}

/*----------------------------------------------------------------------------------*/

static void cron_impex_merge_monthes()
{
  for ( idx_line = 0; idx_line < nb_lines_cron_struct; idx_line++ )
   {
     /* Retrieving the "key" hhmm_id, days_list, grp_action */

     the_hh_id = line_cron_struct[idx_line].hhmi_id;
     the_day_month = line_cron_struct[idx_line].day_month;
     grp_action = line_cron_struct[idx_line].action;

     /* Encode the first month. month represents
       now the monthes as a bit-array 0xDNOSAJJMAMFJ */

     the_month = 0x1 << line_cron_struct[idx_line].month;

     for ( idx_same = idx_line + 1;
           idx_same < nb_lines_cron_struct; idx_same++ )
       {
         if ( ( the_hh_id == line_cron_struct[idx_same].hhmi_id ) &&
              ( grp_action == line_cron_struct[idx_same].action ) &&
              ( the_day_month == line_cron_struct[idx_same].day_month ) )
          {
            /* Combine the month with the first line encountered */

            the_month |= 0x1 << line_cron_struct[idx_same].month;

            /* The line can be removed with the shift */

            for ( idx_shift = idx_same;
                  idx_shift < nb_lines_cron_struct - 1; idx_shift++ )

            line_cron_struct[idx_shift] = line_cron_struct[idx_shift + 1];

            nb_lines_cron_struct -= 1;
            idx_same--;
          }
       }
     line_cron_struct[idx_line].month = the_month;
   }
}

/*----------------------------------------------------------------------------------*/

static void cron_impex_line_cron( gint the_line )
{
  /*   For remembering, structure of one crontab line is:
         min hour day_month month day_week action
   */

  guint idx_bit, mask;
  gboolean first = TRUE;
  gchar scratch[32];

  /* first is the timestamp, forced on 2 digits format (padding) */

  sprintf( str_cron_line,"%.2d %.2d",
           table_horaires[line_cron_struct[the_line].hhmi_id].minutes,
           table_horaires[line_cron_struct[the_line].hhmi_id].heures
         );

  /* Day of month field */

  for ( idx_bit = 1; idx_bit <= 31; idx_bit++ )
     {
       mask = 0x1 << idx_bit;
       if ( ( line_cron_struct[idx_line].day_month & mask ) != 0 )
         {
           if ( first )
             {
               sprintf( scratch, " %.2d", idx_bit );
               first = FALSE;
             }
           else
             sprintf( scratch, ",%.2d", idx_bit );

           strcat( str_cron_line, scratch );
         }
     }

  /* Month field */

  first = TRUE;

  for ( idx_bit = 1; idx_bit <= 12; idx_bit++ )
     {
       mask = 0x1 << idx_bit;
       if ( ( line_cron_struct[idx_line].month & mask ) != 0 )
         {
           if ( first )
             {
               sprintf( scratch, " %.2d", idx_bit );
               first = FALSE;
             }
           else
             sprintf( scratch, ",%.2d", idx_bit );

           strcat( str_cron_line, scratch );
         }
     }

  /* Action to perform. We don't care of the day of week */

  sprintf( scratch, " * /usr/bin/chauffage %d\n", line_cron_struct[idx_line].action );
  strcat( str_cron_line, scratch );
}

/*----------------------------------------------------------------------------------*/

gint cron_impex_export_crontab( day_schedule_t (*schedule), const gchar * file_name )
{
  the_schedule = schedule;

  /* Inits and memory allocations */

  nb_lines_cron_struct = 0;
  line_cron_struct = g_new( cron_flat_line_t, 1024 ); /* exits if fails */
  nb_cron_struct_allocated = 1024;

  /* Remove crontab file if exists and then prepare a new one */

  remove(file_name);
  file_cron_desc = open( file_name, O_WRONLY | O_CREAT | O_NDELAY,
                           S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );

  if ( file_cron_desc < 0 ) /* What is with the rights access... ? */
    {
      g_free( line_cron_struct );
      return CRON_IMPEX_ERR_OPEN_FILE;
    }

  /* Writing the header of the crontab. */

  retour_fonctions = write( file_cron_desc, cron_en_tete, strlen( cron_en_tete ) );

  if ( retour_fonctions < 0 ) /*  Why can't we write into the cron file ? */
    {
      close( file_cron_desc );
      g_free( line_cron_struct );
      return CRON_IMPEX_ERR_WRITE_FILE;
    }

  cron_impex_expand_cron_struct(); /* Expand the days of the schedule */

  cron_impex_merge_group_actions(); /* Encode the actions with tracking */

  cron_impex_merge_time_stamp(); /* Merge the actions within a day */

  cron_impex_merge_days_within_month(); /* ...that have the same actions */

  cron_impex_merge_monthes(); /* ...that have the same sub-schedule */

  /*cron_impex_print_cron();*/

  for ( idx_line = 0; idx_line < nb_lines_cron_struct; idx_line++ )
    {
      cron_impex_line_cron( idx_line );

      retour_fonctions = write( file_cron_desc,
                                str_cron_line, strlen( str_cron_line ) );

      if ( retour_fonctions < 0 )
        {
          close(file_cron_desc);
          g_free( line_cron_struct );
          return CRON_IMPEX_ERR_WRITE_FILE;
        }
    }

  g_free( line_cron_struct );

  close(file_cron_desc);
  return CRON_IMPEX_OK;
}
