/***************************************************************************
                          cron_import.c  -  description
                             -------------------
    begin                : Sat Oct 28 2000
    copyright            : (C) 2000 by Dominique Laigle
    email                : lainal@gmx.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef CRON_IMPEX_H
#include <cron_impex.h>
#endif

#include <time.h>

#ifndef _CRON_FLAT_LINE_T
#define _CRON_FLAT_LINE_T
typedef struct {
                 guint       month;
                 guint       day_month;
                 guchar      hhmi_id;
                 guchar      group_id;
                 guchar      action;
               } cron_flat_line_t;
#endif

#ifndef GRP_NONE
#define GRP_NONE 0xff
#endif

#ifndef LINE_MONTH_DONE
#define LINE_MONTH_DONE  0x1f
#define LINE_DAY_DONE    0x2f
#define LINE_TIME_SORTED 0x3f
#endif


static gint   idx_group, idx_line, idx_ins, idx_same, idx_shift,
              file_cron_desc, retour,
              i, j;
static guint  nb_lines_cron_struct, nb_cron_struct_allocated,
              the_month, the_day_month, the_group, mois, jours;
static guchar str_cron_line[256], group_definition_mask[MAX_GROUPES],
              heures, minutes,gp_actions;

static day_schedule_t   *the_schedule;
static cron_flat_line_t *line_cron_struct, buffer_line_1;

static const horaire_t table_horaires[97]={
{ 0, 0 }, { 0, 15 }, { 0, 30 }, { 0, 45 }, { 1, 0 }, { 1, 15 }, { 1, 30 }, { 1, 45 },
{ 2, 0 }, { 2, 15 }, { 2, 30 }, { 2, 45 }, { 3, 0 }, { 3, 15 }, { 3, 30 }, { 3, 45 },
{ 4, 0 }, { 4, 15 }, { 4, 30 }, { 4, 45 }, { 5, 0 }, { 5, 15 }, { 5, 30 }, { 5, 45 },
{ 6, 0 }, { 6, 15 }, { 6, 30 }, { 6, 45 }, { 7, 0 }, { 7, 15 }, { 7, 30 }, { 7, 45 },
{ 8, 0 }, { 8, 15 }, { 8, 30 }, { 8, 45 }, { 9, 0 }, { 9, 15 }, { 9, 30 }, { 9, 45 },
{ 10, 0 }, { 10, 15 }, { 10, 30 }, { 10, 45 }, { 11, 0 }, { 11, 15 }, { 11, 30 }, { 11, 45 },
{ 12, 0 }, { 12, 15 }, { 12, 30 }, { 12, 45 }, { 13, 0 }, { 13, 15 }, { 13, 30 }, { 13, 45 },
{ 14, 0 }, { 14, 15 }, { 14, 30 }, { 14, 45 }, { 15, 0 }, { 15, 15 }, { 15, 30 }, { 15, 45 },
{ 16, 0 }, { 16, 15 }, { 16, 30 }, { 16, 45 }, { 17, 0 }, { 17, 15 }, { 17, 30 }, { 17, 45 },
{ 18, 0 }, { 18, 15 }, { 18, 30 }, { 18, 45 }, { 19, 0 }, { 19, 15 }, { 19, 30 }, { 19, 45 },
{ 20, 0 }, { 20, 15 }, { 20, 30 }, { 20, 45 }, { 21, 0 }, { 21, 15 }, { 21, 30 }, { 21, 45 },
{ 22, 0 }, { 22, 15 }, { 22, 30 }, { 22, 45 }, { 23, 0 }, { 23, 15 }, { 23, 30 }, { 23, 45 },
{ 23, 59 } /* should be 0:0 the day after... hence this small trick ! */
};

/*----------------------------------------------------------------------------------*/
/* DEBUG PURPOSE ONLY */
/*
static void cron_impex_print_cron()
{
  gint idx;
  printf("\n%d Lines\n",nb_lines_cron_struct);
  for ( idx = 0; idx < nb_lines_cron_struct; idx++ )
    printf("hh:%d mm:%d - M:%u D:%u G:%d A:%d\n",
        table_horaires[line_cron_struct[idx].hhmi_id].heures,
				table_horaires[line_cron_struct[idx].hhmi_id].minutes,
				line_cron_struct[idx].month,
				line_cron_struct[idx].day_month,
				line_cron_struct[idx].group_id,
				line_cron_struct[idx].action);
}
*/
/*----------------------------------------------------------------------------------*/

static void cron_impex_load_cron_line()
{
  gchar **the_tokens, /* fields separated by space (cron items) */
        **sub_tokens; /* fields separated by coma (within monthes, days) */
  gint nb_tokens, nb_sub_tokens, idx_str;

  nb_tokens = nb_sub_tokens = idx_str = 0;

  /* Split the string comming from the line of the crontab */

  the_tokens = g_strsplit( str_cron_line," ", 0 );

  g_return_if_fail( the_tokens != NULL ); /* what is that line ? */

  /* count the tokens within a cron line : it must be 7 */

  for ( nb_tokens = 0; the_tokens[nb_tokens] != NULL; nb_tokens++ );

  g_return_if_fail( nb_tokens == 7 ); /* what is that line ? */

  /* In case we have to expand the structure */

  if ( nb_lines_cron_struct >= nb_cron_struct_allocated - 16 )
   {
     line_cron_struct = g_renew( cron_flat_line_t, line_cron_struct,
                               nb_cron_struct_allocated + 1024 );
     nb_cron_struct_allocated += 1024;
   }

  /* Retrieving the time-stamp and action : always unique */

  minutes = atoi( the_tokens[0] );
  heures = atoi( the_tokens[1] );
  gp_actions = atoi( the_tokens[6] );

  line_cron_struct[nb_lines_cron_struct].hhmi_id = heures * 4 + minutes / 15;
  line_cron_struct[nb_lines_cron_struct].action = gp_actions;
  line_cron_struct[nb_lines_cron_struct].group_id = 0;

  /* case of the days : there can be more separated by coma */

  sub_tokens = g_strsplit( the_tokens[2],",", 0 );

  for ( nb_sub_tokens = 0;
        sub_tokens[nb_sub_tokens] != NULL; nb_sub_tokens++ );

  jours = 0x0;
  for ( i = 0; i < nb_sub_tokens; i++ )
    jours |= 1 << atoi( sub_tokens[i] );

  line_cron_struct[nb_lines_cron_struct].day_month = jours;
  g_strfreev( sub_tokens );

  /* case of the monthes : there can be more separated by coma */

  sub_tokens = g_strsplit( the_tokens[3],",", 0 );

  for ( nb_sub_tokens = 0;
        sub_tokens[nb_sub_tokens] != NULL; nb_sub_tokens++ );

  mois = 0x0;
  for ( i = 0; i < nb_sub_tokens; i++ )
    mois |= 1 << atoi( sub_tokens[i] );

  line_cron_struct[nb_lines_cron_struct].month = mois;
  g_strfreev( sub_tokens );

  g_strfreev( the_tokens );
  nb_lines_cron_struct++;
}

/*-----------------------------------------------------------------------------------------------*/

static void cron_impex_expand_day_month()
{
  gboolean inserted;

  /* In case we must expand the structure - we may add up to 362 lines */

  if ( nb_lines_cron_struct >= nb_cron_struct_allocated - 400 )
   {
     line_cron_struct = g_renew( cron_flat_line_t, line_cron_struct,
                               nb_cron_struct_allocated + 1024 );
     nb_cron_struct_allocated += 1024;
   }

  /* We start by the monthes. The inserts have to respect the upgoing
     time sorting (we will complete the sort once all the lines are
     imported into the structure). */

  for ( idx_line = 0; idx_line < nb_lines_cron_struct; idx_line++ )
   {
     buffer_line_1 = line_cron_struct[idx_line];
     inserted = FALSE;

     for ( i = 1; i <= 12; i++ ) /* scan of the monthes */
      {
        mois = 0x1 << i;
        if ( ( ( mois & line_cron_struct[idx_line].month ) != 0 )
            && ( line_cron_struct[idx_line].group_id < LINE_MONTH_DONE ) )
         {
           buffer_line_1.month = i;
           buffer_line_1.group_id = LINE_MONTH_DONE; /* as "done" flag */

           /* shift the lines to make room */

           for ( idx_shift = nb_lines_cron_struct;
                 idx_shift >= idx_line; idx_shift-- )
             line_cron_struct[idx_shift+1] = line_cron_struct[idx_shift];

           /* insert the new expanded line */

           line_cron_struct[idx_line] = buffer_line_1;

           inserted = TRUE;
           idx_line++; /* in order to stay on the current line */
           nb_lines_cron_struct++; /* we have one more */
         }
      }

     /* Now we have to remove the actual line wich we are finished with */

     if ( inserted == TRUE )
      {
        for ( idx_shift = idx_line;
              idx_shift < nb_lines_cron_struct - 1; idx_shift++ )
          line_cron_struct[idx_shift] = line_cron_struct[idx_shift+1];

        nb_lines_cron_struct--;
        idx_line--; /* in order that it points to the next line */
      }
   }

  /* Now, expansion of the days within a month */

  for ( idx_line = 0; idx_line < nb_lines_cron_struct; idx_line++ )
   {
     buffer_line_1 = line_cron_struct[idx_line];
     inserted = FALSE;

     for ( i = 1; i <= 31; i++ ) /* scan of the days of month */
      {
        jours = 0x1 << i;
        if ( ( ( jours & line_cron_struct[idx_line].day_month ) != 0 )
            && ( line_cron_struct[idx_line].group_id < LINE_DAY_DONE ) )
         {
           buffer_line_1.day_month = i;
           buffer_line_1.group_id = LINE_DAY_DONE; /* as "done" flag */

           /* shift the lines to make room */

           for ( idx_shift = nb_lines_cron_struct;
                 idx_shift >= idx_line; idx_shift-- )
             line_cron_struct[idx_shift+1] = line_cron_struct[idx_shift];

           /* insert the new expanded line */

           line_cron_struct[idx_line] = buffer_line_1;

           inserted = TRUE;
           idx_line++; /* in order to stay on the current line */
           nb_lines_cron_struct++; /* we have one more */
         }
      }

     /* Now we have to remove the actual line wich we are finished with */

     if ( inserted == TRUE )
      {
        for ( idx_shift = idx_line;
              idx_shift < nb_lines_cron_struct - 1; idx_shift++ )
          line_cron_struct[idx_shift] = line_cron_struct[idx_shift+1];

        nb_lines_cron_struct--;
        idx_line--; /* in order that it points to the next line */
      }
   }
}

/*-----------------------------------------------------------------------------------------------*/

void cron_impex_sort_lines()
{
  /* The time is already sorted. So we just have to sort first the days,
     and then the monthes */

  for ( idx_line = 0; idx_line < nb_lines_cron_struct; idx_line++ )
   {
     /* Seek the insertion point. The scanned lines are already sorted */

     the_day_month = line_cron_struct[idx_line].day_month;

     for ( idx_ins = idx_line;
           ( idx_ins < nb_lines_cron_struct - 1 ) &&
           ( line_cron_struct[idx_ins].day_month == the_day_month );
           idx_ins++ )
       idx_line = idx_ins; /* No use to scan ahead up to here */

     if ( idx_ins < nb_lines_cron_struct - 1) /* then new insertion point */
      {
        /* scan to find the same days forward */

        for ( idx_same = idx_ins + 1; idx_same < nb_lines_cron_struct; idx_same++ )
          if ( line_cron_struct[idx_same].day_month == the_day_month )
          {
            buffer_line_1 = line_cron_struct[idx_same];

            /* shift downward to make room */

            for ( idx_shift = idx_same;
                  idx_shift > idx_ins; idx_shift-- )
              line_cron_struct[idx_shift] = line_cron_struct[idx_shift-1];

            /* Insert the line at the right place */

            line_cron_struct[idx_ins] = buffer_line_1;

            idx_ins++; /* The Ins Pt stays on the foreign day */
          }
      }
   }

  /* Sort the monthes */

  for ( idx_line = 0; idx_line < nb_lines_cron_struct; idx_line++ )
   {
     /* Seek the insertion point. The scanned lines are already sorted */

     the_month = line_cron_struct[idx_line].month;

     for ( idx_ins = idx_line;
           ( idx_ins < nb_lines_cron_struct - 1 ) &&
           ( line_cron_struct[idx_ins].month == the_month ); idx_ins++ )
       idx_line = idx_ins; /* No use to scan ahead up to here */

     if ( idx_ins < nb_lines_cron_struct - 1) /* then new insertion point */
      {
        /* scan to find the same monthes forward */

        for ( idx_same = idx_ins + 1; idx_same < nb_lines_cron_struct; idx_same++ )
          if ( line_cron_struct[idx_same].month == the_month )
          {
            buffer_line_1 = line_cron_struct[idx_same];

            /* shift downward to make room */

            for ( idx_shift = idx_same;
                  idx_shift > idx_ins; idx_shift-- )
              line_cron_struct[idx_shift] = line_cron_struct[idx_shift-1];

            /* Insert the line at the right place */

            line_cron_struct[idx_ins] = buffer_line_1;

            idx_ins++; /* The Ins Pt stays on the foreign month */
          }
      }
   }
}

/*-----------------------------------------------------------------------------------------------*/

static int cron_impex_expand_groups()
{
  gchar    env_valeur, *gp_var_env_nom, *gp_var_env_val,
          *gp_id = g_new0( gchar, 3 );
  guchar   all_groups, avant, apres, transition, nb_grp,
           groupe[MAX_GROUPES] = { 0 };
  gboolean env_ok;

  /* Retrieve group definition from the environement. If no environment
     is set : good by. */

  for ( idx_group = 0; idx_group < MAX_GROUPES; idx_group++ )
    {
       g_snprintf( gp_id, 2, "%d", idx_group );
       gp_var_env_nom = g_strconcat( "HEATHER_GP_", gp_id, NULL );

       gp_var_env_val = getenv( gp_var_env_nom );

       g_return_val_if_fail( gp_var_env_val != NULL,
                             CRON_IMPEX_ERR_CYCLE_DEF );

       env_valeur = ( guchar ) strtol( gp_var_env_val, NULL, 0 );
       group_definition_mask[idx_group] = env_valeur;

       g_free( gp_var_env_nom );
    }

  g_free( gp_id );

  /* Test the coherence of the environment : all the bit assignments
     must be unique within the 4 groups. */

  all_groups = group_definition_mask[0];
  env_ok = TRUE;

  for ( i = 1; i < MAX_GROUPES; i++ )
    if ( ( all_groups & group_definition_mask[i] ) !=0 )
     {
       env_ok = FALSE;
       break;
     }
    else all_groups |= group_definition_mask[i];

  g_return_val_if_fail( env_ok == TRUE, CRON_IMPEX_ERR_CYCLE_DEF); /* cycle definition is broken */

  /* Ok : let's go. We expand the groups by determining the transitions
   * between the actual line and the former ( hence the importance of the
   * time sorting ).
   */

  /* In case we must expand the structure - we may magnify the size by 4.
     As it is the last time we should have to expand the structure, we do
     it close to the needs. */

  if ( nb_lines_cron_struct >= nb_cron_struct_allocated / 4 )
   {
     i = 4 + nb_cron_struct_allocated - ( nb_lines_cron_struct * 4 );
     j = nb_cron_struct_allocated + i;

     line_cron_struct = g_renew( cron_flat_line_t, line_cron_struct, j );
     nb_cron_struct_allocated += i;
   }

  avant = 0;
  for ( idx_line = 0; idx_line < nb_lines_cron_struct; idx_line++ )
   {
     buffer_line_1 = line_cron_struct[idx_line];
     apres = buffer_line_1.action;
     transition = ( avant & ~apres ) | ( ~avant & apres );

     /* Retrieve the involved groups */

     nb_grp = 0;
     for ( i = 0; i < MAX_GROUPES; i++ )
       if ( ( transition & group_definition_mask[i] ) > 0 )
         groupe[nb_grp++] = i;

     g_return_val_if_fail( nb_grp > 0, CRON_IMPEX_ERR_DATA_PH_1  );

     /* shift downward to make room. */

     if ( nb_grp > 1 )
      {
        nb_lines_cron_struct += nb_grp - 1;
        for ( idx_shift = nb_lines_cron_struct;
              idx_shift >= idx_line + nb_grp - 1; idx_shift-- )
           line_cron_struct[idx_shift] = line_cron_struct[idx_shift-nb_grp+1];
      }

     /* Scanning the groups to expand the line */

     for ( i = 0; i < nb_grp; i++ )
      {
        idx_group = groupe[i];

        buffer_line_1.group_id = idx_group;
        buffer_line_1.action =
               ( transition & apres & group_definition_mask[idx_group] ) ?
                                           CRON_IMPEX_ON : CRON_IMPEX_OFF;

        line_cron_struct[idx_line] = buffer_line_1;

        idx_line++;
      }
     idx_line--;
     avant = apres;
   }

  /* Sort the groups within the day of month. Then the action should be
     alternatively ON-OFF-ON-OFF .... */

  for ( idx_line = 0; idx_line < nb_lines_cron_struct; idx_line++ )
   {
     /* Seek the insertion point. The scanned lines are already sorted */

     the_month = line_cron_struct[idx_line].month;
     the_day_month = line_cron_struct[idx_line].day_month;
     the_group = line_cron_struct[idx_line].group_id;

     for ( idx_ins = idx_line;
           ( idx_ins < nb_lines_cron_struct - 1 ) &&
           ( line_cron_struct[idx_ins].month == the_month ) &&
           ( line_cron_struct[idx_ins].day_month == the_day_month ) &&
           ( line_cron_struct[idx_ins].group_id == the_group );
           idx_ins++ )
       idx_line = idx_ins; /* No use to scan ahead up to here */

     if ( idx_ins < nb_lines_cron_struct - 1) /* then new insertion point */
      {
        /* scan to find the same monthes,days,group_id forward */

        for ( idx_same = idx_ins + 1; idx_same < nb_lines_cron_struct; idx_same++ )
          if ( ( line_cron_struct[idx_same].month == the_month ) &&
               ( line_cron_struct[idx_same].day_month == the_day_month ) &&
               ( line_cron_struct[idx_same].group_id == the_group ) )
          {
            buffer_line_1 = line_cron_struct[idx_same];

            /* shift downward to make room */

            for ( idx_shift = idx_same;
                  idx_shift > idx_ins; idx_shift-- )
              line_cron_struct[idx_shift] = line_cron_struct[idx_shift-1];

            /* Insert the line at the right place */

            line_cron_struct[idx_ins] = buffer_line_1;

            idx_ins++; /* The Ins Pt stays on the foreign month,day.gr_id */
          }
      }
   }
  return CRON_IMPEX_OK;
}

/*-----------------------------------------------------------------------------------------------*/

static int cron_impex_insert_schedule( )
{
  day_schedule_t  schedule_line;
  gint            idx_cycle, idx_schedule;
  struct tm *actuel;
  time_t    secondes;
  GDate     *the_date;

  /* We pick up the actual date, just to get the current year.
  That means, running the import of crontab within the Sylvester
  night yells some undefined behaviors !!! */

  secondes = time(NULL);
  actuel = localtime( &secondes );
  the_date = g_date_new_dmy( actuel->tm_mday,
                             actuel->tm_mon + 1,
                             actuel->tm_year + 1900 );

  for ( idx_line = 0; idx_line < nb_lines_cron_struct - 1; idx_line++ )
   {
     the_month = line_cron_struct[idx_line].month;
     the_day_month = line_cron_struct[idx_line].day_month;
     the_group = line_cron_struct[idx_line].group_id;

     g_date_set_day( the_date, the_day_month );
     g_date_set_month( the_date, the_month );

     idx_schedule = g_date_day_of_year( the_date ) - 1;
     schedule_line = the_schedule[idx_schedule];
     schedule_line.date = *the_date;

     for ( idx_ins = idx_line + 1;
           ( idx_ins < nb_lines_cron_struct ) &&
           ( line_cron_struct[idx_ins].month == the_month ) &&
           ( line_cron_struct[idx_ins].day_month == the_day_month ) &&
           ( line_cron_struct[idx_ins].group_id == the_group );
           idx_ins++ );

     idx_ins--; /* All in all, we are one step too far */

     if ( idx_ins >= idx_line + 1 ) /* we isolated a group */
      {
        idx_cycle = 0;

        for ( i = idx_line; i < idx_ins; i+= 2 )
         {
           g_return_val_if_fail( line_cron_struct[i].action == CRON_IMPEX_ON,
                                 CRON_IMPEX_ERR_CYCLE_ALIASING );
           g_return_val_if_fail( line_cron_struct[i+1].action == CRON_IMPEX_OFF,
                                 CRON_IMPEX_ERR_CYCLE_NO_END );

           /* Insert the beginning of the cycle */

           schedule_line.day_program.day_cycles[the_group][idx_cycle].heure_debut.heures =
                        table_horaires[line_cron_struct[i].hhmi_id].heures;
           schedule_line.day_program.day_cycles[the_group][idx_cycle].heure_debut.minutes =
                       table_horaires[line_cron_struct[i].hhmi_id].minutes;

           /* Insert the end of the cycle */

           schedule_line.day_program.day_cycles[the_group][idx_cycle].heure_fin.heures =
                        table_horaires[line_cron_struct[i+1].hhmi_id].heures;
           schedule_line.day_program.day_cycles[the_group][idx_cycle].heure_fin.minutes =
                       table_horaires[line_cron_struct[i+1].hhmi_id].minutes;

           idx_cycle++;
         }
        schedule_line.day_program.nb_cycles[the_group] = idx_cycle;
        the_schedule[idx_schedule] = schedule_line;
        idx_line = idx_ins;
      }
   }
  return CRON_IMPEX_OK;
}

/*-----------------------------------------------------------------------------------------------*/

gint cron_impex_import_crontab( const gchar * file_name, day_schedule_t (*schedule) )
{
  gboolean encore;
  guint idx_char;

  the_schedule = schedule;

  nb_lines_cron_struct = 0;
  line_cron_struct = g_new( cron_flat_line_t, 1024 ); /* exits if fails */

  nb_cron_struct_allocated = 1024;

  /* Ouverture en lecture du fichier crontab */

  file_cron_desc = open( file_name, O_RDONLY | O_NDELAY );

  if ( file_cron_desc < 0 ) /* Yaurait pas un problme de droits d'accs ou d'existence...?     */
      return CRON_IMPEX_ERR_OPEN_FILE;

  /* Position de la tte de lecture en dbut de fichier */

  lseek( file_cron_desc, 0, SEEK_SET ); /* Positionnement de la tte de lecture au dbut  */

  /* Boucle de parcour du fichier crontab */

  encore = TRUE;
  idx_char = 0; /* Attention : on ne vrifie pas qu'on prend plus de 256 octets sans "\n" */

  do {
       retour = read( file_cron_desc, (void*) &str_cron_line[idx_char], 1);
       if ( retour > 0 )
        {
          if ( str_cron_line[idx_char] == '\n') /* On a isol une ligne...                                    */
           {
             str_cron_line[++idx_char] = 0x0;

             cron_impex_load_cron_line(); /* import the line */
             cron_impex_expand_day_month();

             idx_char = 0; /* On recommence pour une nouvelle ligne */
           }
          else idx_char++; /* tant que pas ligne complte, caractres suivant */
        }
       else encore = FALSE; /* Fin du fichier ou erreur de lecture : on s'en va  */
     }
  while(encore);

  cron_impex_sort_lines(); /* Necessary for the right group expansion */

  retour = cron_impex_expand_groups();

  /*cron_impex_print_cron();*/

  retour = cron_impex_insert_schedule( );

  close( file_cron_desc );
  return CRON_IMPEX_OK;
}

/*-----------------------------------------------------------------------------------------------*/

gchar *cron_impex_verbose_error_message( gint error_msg )
{
switch ( error_msg )
  {
    case CRON_IMPEX_OK : return "Ok";break;
    case CRON_IMPEX_ERR_ALLOC_MEM : return "memory allocation";break;
    case CRON_IMPEX_ERR_DATA_PH_1 : return "data consistency";break;
    case CRON_IMPEX_ERR_OPEN_FILE : return "opening file";break;
    case CRON_IMPEX_ERR_WRITE_FILE : return "writing file";break;
    case CRON_IMPEX_ERR_CYCLE_ALIASING : return "cycles aliasing";break;
    case CRON_IMPEX_ERR_CYCLE_NO_END : return "non-ended cycle";break;
    case CRON_IMPEX_ERR_TOO_MUCH_CYCLE : return "too much cycles";break;
    default:return "unknown error";break;
  }
}