/***************************************************************************
                          groupadmin.c  -  description
                             -------------------
    begin                : Sat Apr 14 2001
    copyright            : (C) 2001 by Dominique Laigle
    email                : lainal@gmx.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <groupadmin.h>
#include <gtkflagbinaire.h>
#include <gtk/gtkradiobutton.h>
#include <libgnomeui/gnome-dialog.h>
#include <libgnomeui/gnome-stock.h>

/* Static data which constitute the private data of the widget */

/* Signals this widget sends */

enum {
       SIGNAL_NEW_GROUP_SELECT,
       SIGNAL_NEW_GROUP_DEFINED,
       LAST_SIGNAL
     };

static gint groupe_admin_signals[LAST_SIGNAL] = { 0 };

static GtkHBoxClass *parent_class = NULL;

static gint i, j;
static GtkWidget *button_container,
                 *the_dialog;
static gchar      buffer[2];

static GtkWidget *bit_radio_button[4][8]; /* must be visible to the call-back */
static guchar     grp_def_mask_buffer[4]; /* copied in attribute only if dialog = OK */

static gchar *gp_var_env_nom, *gp_var_env_val, *gp_id;
static gboolean *using_default;

/* Forward references */

static void slot_group_btn_toggled( GtkWidget *, GroupeAdmin * );
static void slot_def_bit_toggled( GtkWidget * );
static void slot_def_btn_clicked( GtkWidget *, GroupeAdmin * );

static void groupe_admin_class_init(GroupeAdminClass *);
static void groupe_admin_init( GroupeAdmin * );

/*-------------------------------------------------------------------------*/

GtkType groupe_admin_get_type( void )
{
  GtkType groupe_admin_type = 0;

  if ( !groupe_admin_type )
    {
      static const GtkTypeInfo groupe_admin_info =
      {
	      "GroupeAdmin",
	      sizeof (GroupeAdmin),
	      sizeof (GroupeAdminClass),
	      (GtkClassInitFunc)  groupe_admin_class_init,
	      (GtkObjectInitFunc) groupe_admin_init,
	      (GtkArgSetFunc) NULL,
	      (GtkArgGetFunc) NULL,
      };

      /* On remonte le type de la classe dont on herite */

      groupe_admin_type = gtk_type_unique( GTK_TYPE_HBOX, &groupe_admin_info );
    }
  return groupe_admin_type;
}

/*-------------------------------------------------------------------------*/

void groupe_admin_class_init(GroupeAdminClass *klass)
{
  GtkObjectClass *object_class;

  object_class = (GtkObjectClass*) klass;

  parent_class = gtk_type_class( GTK_TYPE_WIDGET );

  /* Signals registration */

  groupe_admin_signals[SIGNAL_NEW_GROUP_SELECT] =
    gtk_signal_new( "new_group_select",
                     GTK_RUN_FIRST,
                     object_class->type,
                     GTK_SIGNAL_OFFSET( GroupeAdminClass,
                                        new_group_select),
                     gtk_signal_default_marshaller, GTK_TYPE_NONE, 0 );

  groupe_admin_signals[SIGNAL_NEW_GROUP_DEFINED] =
    gtk_signal_new( "new_group_defined",
                     GTK_RUN_FIRST,
                     object_class->type,
                     GTK_SIGNAL_OFFSET( GroupeAdminClass,
                                        new_group_defined),
                     gtk_signal_default_marshaller, GTK_TYPE_NONE, 0 );

  gtk_object_class_add_signals( object_class,
                                groupe_admin_signals, LAST_SIGNAL );

  /* Default handlers not set to null for debug */

  klass->new_group_select = NULL;
  klass->new_group_defined = NULL;
}

/*-------------------------------------------------------------------------*/

GtkWidget* groupe_admin_new( gboolean* addr_defaut_flag )
{
  using_default = addr_defaut_flag;
  return GTK_WIDGET( gtk_type_new( groupe_admin_get_type() ) );
}

/*-------------------------------------------------------------------------*/

void groupe_admin_init( GroupeAdmin *the_groupe_admin )
{
  GtkWidget *the_widget, *the_box;
  gchar* titres[4] = { "groupe 1", "groupe 2", "groupe 3", "groupe 4" };
  gboolean env_ok;
  guchar all_groups, env_valeur;

  *using_default = FALSE; /* this flag resides in the caller's space */

  the_widget = GTK_WIDGET(the_groupe_admin);

  /* Drawing policy = INDIVIDUAL SUBWIDGET */

  the_box = gtk_vbox_new( TRUE, 0 );
  gtk_container_add( GTK_CONTAINER(the_groupe_admin), the_box );
  gtk_widget_show( the_box );

  /*
   * Creation of all the subwidgets. For composite widget, steps are:
   * - create an instance of the subwidget
   * - package it within a container that belongs to the widget
   * - make it visible (up to the policy: if individual drawing)
   */

  /* hbox for the radio buttons and the def push button */

  button_container = gtk_hbox_new( TRUE, 0 );
  gtk_box_pack_start( GTK_BOX( the_box ),
                      button_container,
                      FALSE, TRUE, 1 );
  gtk_widget_show( button_container );

  /* 4 radio_buttons. The first creates the group and hence has to
     be out of the creation loop */

  the_groupe_admin->group_button[0] =
     gtk_radio_button_new_with_label( NULL, "1" );

  gtk_box_pack_start( GTK_BOX( button_container ),
                      the_groupe_admin->group_button[0],
                      FALSE, TRUE, 20 );
  gtk_widget_show( the_groupe_admin->group_button[0] );

  gtk_signal_connect( GTK_OBJECT( the_groupe_admin->group_button[0] ),
                      "toggled",
		                  GTK_SIGNAL_FUNC(slot_group_btn_toggled),
		                  the_groupe_admin );

  for ( i = 1; i < MAX_GROUPES; i++ )
    {
      g_snprintf( buffer, 2, "%d", i + 1 );
      the_groupe_admin->group_button[i] =
         gtk_radio_button_new_with_label_from_widget(
                       GTK_RADIO_BUTTON( the_groupe_admin->group_button[0] ),
                       buffer );
      gtk_box_pack_start( GTK_BOX( button_container ),
                          the_groupe_admin->group_button[i],
                          FALSE, TRUE, 20 );
      gtk_widget_show( the_groupe_admin->group_button[i] );

      gtk_signal_connect( GTK_OBJECT( the_groupe_admin->group_button[i] ),
                          "toggled",
		                      GTK_SIGNAL_FUNC(slot_group_btn_toggled),
		                      the_groupe_admin );
    }

  the_groupe_admin->activ_group = 0;

  /* Group definition push button within the radio button container */

  the_groupe_admin->definition_btn =
                   gtk_button_new_with_label( _("Groupes Def.") );
  gtk_box_pack_start( GTK_BOX( button_container ),
                      the_groupe_admin->definition_btn,
                      FALSE, TRUE, 0 );
  gtk_widget_show( the_groupe_admin->definition_btn );

  gtk_signal_connect( GTK_OBJECT( the_groupe_admin->definition_btn ),
                      "clicked",
                      GTK_SIGNAL_FUNC(slot_def_btn_clicked),
	                    the_groupe_admin );

  /* flagbinaire from the states of the output */

  the_groupe_admin->group_state_flg = gtk_flag_binaire_new( 4, titres );
  gtk_box_pack_start( GTK_BOX( the_box ),
                      the_groupe_admin->group_state_flg,
                      FALSE, TRUE, 1 );
  gtk_widget_show( the_groupe_admin->group_state_flg );

  /* Group definitions. If the environment is not set, then
     the default values are used. */

  gp_id = g_new0( gchar, 3 );

  for ( i = 0; i < MAX_GROUPES; i++ )
    {
       g_snprintf( gp_id, 2, "%d", i );
       gp_var_env_nom = g_strconcat( "HEATHER_GP_", gp_id, NULL );

       gp_var_env_val = getenv( gp_var_env_nom );

       if ( gp_var_env_val == NULL )
          *using_default = TRUE;
       else { /* We let the base as large as possible. See the man page
                 of strtol for more details */
              env_valeur = ( guchar ) strtol( gp_var_env_val, NULL, 0 );
              the_groupe_admin->group_definition_mask[i] = env_valeur;
            }
       g_free( gp_var_env_nom );
    }

  g_free( gp_id ); /* No Need any more */

  /* Checking the coherence of the environment */

  all_groups = the_groupe_admin->group_definition_mask[0];
  env_ok = TRUE;

  for ( i = 1; i < MAX_GROUPES; i++ )
    if ( ( all_groups & the_groupe_admin->group_definition_mask[i] ) !=0 )
     {
       env_ok = FALSE;
       break;
     }
    else all_groups |= the_groupe_admin->group_definition_mask[i];

  g_return_if_fail( env_ok == TRUE ); /* cycle definition is broken */

  if ( *using_default == TRUE )
    {
      /* The main will show a warning dialog box */

      the_groupe_admin->group_definition_mask[0] = DEFAULT_GROUP_1;
      the_groupe_admin->group_definition_mask[1] = DEFAULT_GROUP_2;
      the_groupe_admin->group_definition_mask[2] = DEFAULT_GROUP_3;
      the_groupe_admin->group_definition_mask[3] = DEFAULT_GROUP_4;

      /* Export the default to the environment */

      gp_id = g_new0( gchar, 3 );
      gp_var_env_val = g_new0( gchar, 8 );

      for ( i = 0; i < MAX_GROUPES; i++ )
        {
           g_snprintf( gp_id, 2, "%d", i );
           gp_var_env_nom = g_strconcat( "HEATHER_GP_", gp_id, NULL );

           g_snprintf( gp_var_env_val, 7, "%#.2x",
                       the_groupe_admin->group_definition_mask[i] );

           if ( setenv( gp_var_env_nom, gp_var_env_val, TRUE ) < 0 )
            {
              fprintf( stderr, "prb with environment\n" );

              g_free( gp_var_env_nom );
              g_free( gp_var_env_val );
              g_free( gp_id );

              g_return_if_fail( 1 );
            }
           g_free( gp_var_env_nom );
        }

      g_free( gp_var_env_val );
      g_free( gp_id );
    }

  /* First signal sent from init. Nothing comes from the radio-buttons */

  gtk_signal_emit( GTK_OBJECT(the_groupe_admin),
                   groupe_admin_signals[SIGNAL_NEW_GROUP_SELECT] );
}

/*-------------------------------------------------------------------------*/

static void slot_group_btn_toggled( GtkWidget *the_widget,
                                    GroupeAdmin *the_groupe_admin )
{
  GtkToggleButton *the_button = GTK_TOGGLE_BUTTON(the_widget);

  /* This signal is received twice : one time from the living button
     and one time from the entering one. We just cascade the 2nd one. */

  for ( i = 0;
        ( i < 4 ) &&
        ( GTK_TOGGLE_BUTTON(the_groupe_admin->group_button[i]) != the_button );
        i++
      );

  g_return_if_fail( i < 4 ); /* who sent the signal then ???????? */

  if ( GTK_TOGGLE_BUTTON(the_groupe_admin->group_button[i])->active )
    {
      the_groupe_admin->activ_group = i;

      gtk_signal_emit( GTK_OBJECT(the_groupe_admin),
                       groupe_admin_signals[SIGNAL_NEW_GROUP_SELECT] );
    }
}

/*-------------------------------------------------------------------------*/

static void slot_def_btn_clicked( GtkWidget *the_widget,
                                  GroupeAdmin *the_groupe_admin )
{
  GtkWidget *h_box_bit, *v_box_group_title, *group_title[4],
            *raw_title, *bit_label[8],      *v_box_bit[8];
  gchar     title_buffer[9];

  /* Creation of the definition dialog */

  the_dialog = gnome_dialog_new( _("Dfinition Groupes"),
                                 GNOME_STOCK_BUTTON_OK,
                                 GNOME_STOCK_BUTTON_CANCEL,
                                 NULL
                               );

  h_box_bit = gtk_hbox_new( FALSE, 0 );
  gtk_box_pack_start( GTK_BOX( GNOME_DIALOG(the_dialog)->vbox ),
                      h_box_bit,
                      FALSE, TRUE, 1 );
  gtk_widget_show( h_box_bit );

  /* Packing the groups labels to the first v_box */

  v_box_group_title = gtk_vbox_new( TRUE, 0 );
  gtk_box_pack_start( GTK_BOX( h_box_bit ),
                      v_box_group_title,
                      FALSE, TRUE, 5 );
  gtk_widget_show( v_box_group_title );

  /* Row title */

  raw_title = gtk_label_new( "Bits Sortie" );
  gtk_box_pack_start( GTK_BOX( v_box_group_title ),
                      raw_title,
                      FALSE, TRUE, 1 );
  gtk_widget_show( raw_title );

  /* Groups title */

  for ( j = 0; j < 4; j++ )
    {
      g_snprintf( title_buffer, 9, "Groupe %d", j + 1 );
      group_title[j] = gtk_label_new( title_buffer );

      gtk_box_pack_start( GTK_BOX( v_box_group_title ),
                          group_title[j],
                          FALSE, TRUE, 1 );
      gtk_widget_show( group_title[j] );
    }

  /* Matrix of 8 bits X  4 groups */

  for ( i = 7; i >= 0; i-- )
    {
      /* Creation of the v_box containing the 4 radio buttons */

      v_box_bit[i] = gtk_vbox_new( TRUE, 0 );
      gtk_box_pack_start( GTK_BOX( h_box_bit ),
                      v_box_bit[i],
                      FALSE, TRUE, 1 );
      gtk_widget_show( v_box_bit[i] );

      /* Header with the bit_title */

      g_snprintf( title_buffer, 4, "%d  ", i );
      bit_label[i] = gtk_label_new( title_buffer );

      gtk_box_pack_start( GTK_BOX( v_box_bit[i] ),
                          bit_label[i],
                          FALSE, TRUE, 5 );
      gtk_widget_show( bit_label[i] );

      /* creation of the first radio button from the bit i */

      bit_radio_button[0][i] = gtk_radio_button_new( NULL );
      gtk_box_pack_start( GTK_BOX( v_box_bit[i] ),
                          bit_radio_button[0][i],
                          FALSE, TRUE, 5 );

      /* creation of the 3 other buttons for this group */

      for ( j = 1; j < 4; j++ )
        {
          bit_radio_button[j][i] =
             gtk_radio_button_new_from_widget
                            ( GTK_RADIO_BUTTON( bit_radio_button[0][i] ) );

          gtk_box_pack_start( GTK_BOX( v_box_bit[i] ),
                              bit_radio_button[j][i],
                              FALSE, TRUE, 5 );
        }
    }

  /* Set the state of the check boxes in relation with the current
     group definition mask.  We do this  before the  radio buttons
     are shown for convenience, but this is done before  they  are
     connected to the call back because there is no mask to update
     since we pick-up their state right now.  */

  for ( j = 0; j < 4; j++ )
  {
    grp_def_mask_buffer[j] = the_groupe_admin->group_definition_mask[j];

    for ( i = 0; i < 8; i++ )
      {
        gtk_toggle_button_set_active(
            GTK_TOGGLE_BUTTON( bit_radio_button[j][i] ),
            ( the_groupe_admin->group_definition_mask[j] & ( 1 << i ) ) );

        gtk_widget_show( bit_radio_button[j][i] );

        gtk_signal_connect( GTK_OBJECT( bit_radio_button[j][i] ),
                            "toggled",
		                        GTK_SIGNAL_FUNC(slot_def_bit_toggled),
		                        NULL );
      }
   }

  /* Run the dialog as modal */

  if ( gnome_dialog_run_and_close( GNOME_DIALOG(the_dialog) ) == 0 )
   {
     /* OK button was pressed to close the dialog, so we
        copy the mask buffers into the object attributes*/

     for ( j = 0; j < 4; j++ )
        the_groupe_admin->group_definition_mask[j] = grp_def_mask_buffer[j];

     /* New masks defined : we have to prevent the appli. */

     gtk_signal_emit( GTK_OBJECT(the_groupe_admin),
                      groupe_admin_signals[SIGNAL_NEW_GROUP_DEFINED] );
   }
}

/*-------------------------------------------------------------------------*/

static void slot_def_bit_toggled( GtkWidget * the_widget )
{
  /* this signal is received twice : one time from the living button
     and one time from the entering one. We take both in consideration
     to update the buffer group masks. */

  GtkToggleButton *the_button = GTK_TOGGLE_BUTTON(the_widget);

  i = j = 0;

  while ( ( j < 4 ) &&
          ( GTK_TOGGLE_BUTTON(bit_radio_button[j][i]) != the_button ) )
  {
    i++;
    if ( i == 8 )
     {
       i = 0;
       j++;
     }
  }

  g_return_if_fail( j < 4 ); /* Where the signal come from ??? */

  if ( GTK_TOGGLE_BUTTON(the_button)->active )
    grp_def_mask_buffer[j] |= ( 1 << i );
  else
    grp_def_mask_buffer[j] &= ~( 1 << i );
}