/***************************************************************************
                          gtkcalendrier.c  -  description
                             -------------------
    begin                : Mon Jan 1 2001
    copyright            : (C) 2001 by Dominique Laigle
    email                : lainal@gmx.de

This widget was at the begining intended to extend gtkcalendar. As there were
neighther selection nor copy-paste facilities in gtkcalendar, I decided to
rewrite everything from scratch. The coding style is no quite compliant
with the Gnome Community's one, but I still hope this widget will be usefull.

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <locale.h>
#include <sys/time.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <stdio.h>

#include <gtkcalendrier.h>

#define CASE_DEFAULT_WIDTH  30
#define CASE_DEFAULT_HEIGHT 25

#define BISSEXTILE(an) ((((an % 4) == 0) && ((an % 100) != 0)) || ((an % 400) == 0))

/* Signals sent by this widget */

enum {
       NEW_DATE_COPY,     /* A new date came into the clipboard */
       NEW_DATE_PASTE,    /* New selection as target of the clipboard */
       NEW_DATE_SELECTED, /* A new date was selected */
       LAST_SIGNAL
     };

static gint gtk_calendrier_signals[LAST_SIGNAL] = { 0 };

/* Static data to understand here as PRIVATE DATA of the object */

static gint case_clic[7][9], clic_i, clic_j;
static gint select_semantic;

static gint mois_fin[12] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
GDate *date_tampon;
static gint i, j, k, case_width, case_height, matrix_selected[7][6];

static GdkRectangle rectangle;
static GdkPixmap   *the_pixmap; /* For drawing the widget */
static GdkGC       *the_gc;

static GtkMenuItem *item_copy,
                   *item_paste;
static GtkMenu     *context_menu;

static GtkWidget   *this_widget; /* Yes, I know: it's hard to drop C++...*/

/*-------- Forward References ----------------------------------*/

static GtkWidgetClass *parent_class = NULL;

static void gtk_calendrier_class_init(GtkCalendrierClass *);
static void gtk_calendrier_init(GtkCalendrier *);
static void gtk_calendrier_realize(GtkWidget *);

static void gtk_calendrier_destroy(GtkObject *);

static void gtk_calendrier_size_request(GtkWidget *, GtkRequisition *);
static void gtk_calendrier_size_allocate(GtkWidget *, GtkAllocation *);

static void gtk_calendrier_paint(GtkWidget *, GdkRectangle *);
static void gtk_calendrier_draw(GtkWidget *, GdkRectangle *);
static gint gtk_calendrier_expose(GtkWidget *, GdkEventExpose *);

static gint gtk_calendrier_button_press(GtkWidget *, GdkEventButton *);
static gint gtk_calendrier_button_release(GtkWidget *, GdkEventButton *);

/*-------------------------------------------------------------------------*/

static void select_transpose()
{
  /* Transpose the content of the clipboard into a new selection
     regarding its semantic. */

  GtkCalendrier *calendrier = GTK_CALENDRIER(this_widget);

  g_return_if_fail( calendrier->clip_board.semantique != PAS_DEFINIE ); /* Theorically impossible */

  select_semantic = calendrier->clip_board.semantique;

  switch ( select_semantic )
  {
    case JOURS_MOIS:

      for ( j = 0; j < 6; j++ )
        for ( i = 0; i < 7; i++ )
          {
            if ( atoi( calendrier->jours_mois[i+7*j] ) ==
                 g_date_day( calendrier->clip_board.dates_copy[0] ) )

              matrix_selected[i][j] = 1;

            else matrix_selected[i][j] = 0;
          }
      break;

    case JOURS_SEMAINE:

      for ( i = 0; i < 7; i++ )
         if ( i != g_date_weekday(calendrier->clip_board.dates_copy[0]) - 1 )
             for ( j = 0; j < 6; j++ )
                 matrix_selected[i][j] = 0;
         else
             for ( j = 0; j < 6; j++ )
                 if ( strcmp( calendrier->jours_mois[i+7*j], "0" ) )
                     matrix_selected[i][j] = 1;
      break;

    case LA_SEMAINE:

      for ( i = 0; i < 5; i++ )
         for ( j = 0; j < 6; j++ )
            if ( strcmp( calendrier->jours_mois[i+7*j], "0" ) )
                 matrix_selected[i][j] = 1;
            else matrix_selected[i][j] = 0;

      for ( i = 5; i < 7; i++ )
         for ( j = 0; j < 6; j++ )
            matrix_selected[i][j] = 0;

      break;

    default:/* LE_WEEK_END */

      for ( i = 0; i < 5; i++ )
         for ( j = 0; j < 6; j++ )
            matrix_selected[i][j] = 0;

      for ( i = 5; i < 7; i++ )
         for ( j = 0; j < 6; j++ )
            if ( strcmp( calendrier->jours_mois[i+7*j], "0" ) )
                 matrix_selected[i][j] = 1;
            else matrix_selected[i][j] = 0;

      break;
  }
}

/*-------------------------------------------------------------------------*/

static void select_reset()
{
  /* Set the selected flag matrix to zero */

  for ( j = 0; j < 6; j++ )
    for ( i = 0; i < 7; i++ )
        matrix_selected[i][j] = 0;
}

/*-------------------------------------------------------------------------*/

static void copy_call_back( GtkMenuItem *the_menu_item )
{
  GtkCalendrier *calendrier = GTK_CALENDRIER( this_widget );

  if ( select_semantic == PAS_DEFINIE ) /* No selection, nothing to copy... */
    return;

  k = 0;

  calendrier->clip_board.semantique = select_semantic;

  for ( j = 0; j < 6; j++ )
    for ( i = 0; i < 7; i++ )
      if ( matrix_selected[i][j] != 0 )
        {
          g_date_set_julian( calendrier->clip_board.dates_copy[k],
                             g_date_julian( calendrier->date_interne ) );

          g_date_set_day( calendrier->clip_board.dates_copy[k++],
                          atoi( calendrier->jours_mois[i+7*j] ) );
        }

  /* If anybody is interested that a copy to clipboard has been performed.*/

  calendrier->clip_board.nb_jours_copy = k;

  gtk_signal_emit( GTK_OBJECT(this_widget),
                   gtk_calendrier_signals[NEW_DATE_COPY] );
}

/*-------------------------------------------------------------------------*/

static void paste_call_back( GtkMenuItem *the_menu_item )
{
  /* It is important to notice here that the number of days in the
     copy area doesn't especially match the one in the paste area.
     It is the responsability of the top application to deal with
     those situations and define the policy to conduct. */

  GtkCalendrier *calendrier = GTK_CALENDRIER( this_widget );

  if ( calendrier->clip_board.nb_jours_copy == 0 )
    return; /* clipboard empty : nothing to paste */

  /* semantic mismatch/undefined : theorically impossible, but still...*/

  g_return_if_fail( calendrier->clip_board.semantique != PAS_DEFINIE );
  g_return_if_fail( calendrier->clip_board.semantique == select_semantic );

  /* Everything seems Ok */

  k = 0;

  for ( j = 0; j < 6; j++ )
    for ( i = 0; i < 7; i++ )
      if ( matrix_selected[i][j] != 0 )
        {
          g_date_set_julian( calendrier->clip_board.dates_paste[k],
                             g_date_julian( calendrier->date_interne ) );

          g_date_set_day( calendrier->clip_board.dates_paste[k++],
                          atoi( calendrier->jours_mois[i+7*j] ) );
        }

  calendrier->clip_board.nb_jours_paste = k;

  /* Signal saying that it is time to perform the data paste */

  gtk_signal_emit( GTK_OBJECT(this_widget),
                   gtk_calendrier_signals[NEW_DATE_PASTE] );
}

/*-------------------------------------------------------------------------*/

static void calcule_date_matrice( gchar matrice[42][3], GDate *date_interne )
{
  gint premier_jour, dernier_jour, jour_courant, mois_courant;

  /* Recherche dernier quantime du mois courant*/

  mois_courant = g_date_month(date_interne) - 1;

  if ( ( mois_courant == 1 ) && BISSEXTILE( g_date_year(date_interne) ) )
       dernier_jour = 29;
  else dernier_jour = mois_fin[mois_courant];

  /* Recherche premier jour du mois courant */

  jour_courant = g_date_day(date_interne);
  g_date_set_day( date_interne, 1);
  premier_jour = g_date_weekday(date_interne);
  g_date_set_day( date_interne, jour_courant);

  /* Initialisation de la matrice. Les 14 premiers et 14 derniers jours
     sont  0, ce qui sert aussi de flag de non-affichage...*/

  for ( i = 0; i < 14; i++ )
     {
       sprintf( matrice[i],"0" );
       sprintf( matrice[41-i],"0");
     }

  /* Remplissage avec les quantimes du mois en commenant au premier
     jour de la semaine dtect */

  for ( i = 1; i <= dernier_jour; i++ )
     sprintf( matrice[i + premier_jour - 2], "%d", i );
}

/*-------------------------------------------------------------------------*/

GtkType gtk_calendrier_get_type( void )
{
  static GtkType calendrier_type = 0;

  if ( !calendrier_type )
    {
      static const GtkTypeInfo calendrier_info =
      {
	      "GtkCalendrier",
	      sizeof (GtkCalendrier),
	      sizeof (GtkCalendrierClass),
	      (GtkClassInitFunc)  gtk_calendrier_class_init,
	      (GtkObjectInitFunc) gtk_calendrier_init,
	      (GtkArgSetFunc) NULL,
	      (GtkArgGetFunc) NULL,
      };
      calendrier_type = gtk_type_unique( GTK_TYPE_WIDGET, &calendrier_info );
    }
  return calendrier_type;
}

/*-------------------------------------------------------------------------*/

static void gtk_calendrier_class_init( GtkCalendrierClass *klass )
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;

  object_class = (GtkObjectClass*) klass;
  widget_class = (GtkWidgetClass*) klass;

  parent_class = gtk_type_class( GTK_TYPE_WIDGET );

  widget_class->realize   = gtk_calendrier_realize;

  object_class->destroy   = gtk_calendrier_destroy;

  widget_class->draw         = gtk_calendrier_draw;
  widget_class->expose_event = gtk_calendrier_expose;

  widget_class->size_request  = gtk_calendrier_size_request;
  widget_class->size_allocate = gtk_calendrier_size_allocate;

  widget_class->button_press_event   = gtk_calendrier_button_press;
  widget_class->button_release_event = gtk_calendrier_button_release;

  gtk_calendrier_signals[NEW_DATE_COPY] =
    gtk_signal_new( "new_date_copy",
		                GTK_RUN_FIRST, object_class->type,
		                GTK_SIGNAL_OFFSET( GtkCalendrierClass, new_date_copy ),
		                gtk_signal_default_marshaller, GTK_TYPE_NONE, 0 );

  gtk_calendrier_signals[NEW_DATE_PASTE] =
    gtk_signal_new( "new_date_paste",
		                GTK_RUN_FIRST, object_class->type,
		                GTK_SIGNAL_OFFSET( GtkCalendrierClass, new_date_paste ),
		                gtk_signal_default_marshaller, GTK_TYPE_NONE, 0 );

  gtk_calendrier_signals[NEW_DATE_SELECTED] =
    gtk_signal_new( "new_date_selected",
		                GTK_RUN_FIRST, object_class->type,
		                GTK_SIGNAL_OFFSET( GtkCalendrierClass, new_date_selected ),
		                gtk_signal_default_marshaller, GTK_TYPE_NONE, 0 );

  gtk_object_class_add_signals( object_class,
                                gtk_calendrier_signals,
                                LAST_SIGNAL
                               );

  /* klass->call backs = NULL pour annoncer qui ya pas
     de traitements par defaut  */
  klass->new_date_copy = NULL;
  klass->new_date_paste = NULL;
  klass->new_date_selected = NULL;
}

/*-------------------------------------------------------------------------*/

static void gtk_calendrier_init( GtkCalendrier *calendrier )
{
  struct tm *actuel;
  time_t    secondes;

  this_widget = GTK_WIDGET(calendrier);
  GTK_WIDGET_SET_FLAGS (this_widget, GTK_CAN_FOCUS);

  the_pixmap = NULL;

  /* A l'init, on positione la date interne  la date courante */

  secondes = time(NULL);
  actuel = localtime( &secondes );
  calendrier->date_interne = g_date_new_dmy( actuel->tm_mday,
                                             actuel->tm_mon + 1,
                                             actuel->tm_year + 1900 );

  calendrier->date_courante = g_date_new_dmy( actuel->tm_mday,
                                              actuel->tm_mon + 1,
                                              actuel->tm_year + 1900 );

  /* Calcule de la matrice des jours du mois */

  calcule_date_matrice( calendrier->jours_mois, calendrier->date_interne );

  /* Positionnement des tableaux des jours de la semaine et des mois de
     l'anee. On utilise les fonctions systme pour profiter de l'i18n.
  */

  setlocale(LC_ALL,""); /* C'est pas le boulot de Gtk de faire cela */
  for ( i = 0; i < 7; i++ )
     {
       /* On prfre que la semaine commence un lundi et pas un dimanche
          d'o le dcalage des indices de calcul. */
       secondes = ( i + 4 ) * 86400;
       strftime( calendrier->jours_semaine[i], 4, "%a", gmtime(&secondes) );
     }

  for ( i = 0; i < 12; i++ )
    {
      secondes = i * 2764800;
      strftime( calendrier->mois_annee[i], 10, "%B", gmtime(&secondes) );
    }

  /* Pour le header de la widget, tous les textes sont aligns en hauteur sur
     la base d'un caractre sans jambage ni boucle : on est ainsi sr d'un
     centrage constant en hauteur (on a pris "n") */

  calendrier->hauteur_texte = gdk_char_height( this_widget->style->font, 'n' );

  /* A l'init, aucune case n'a t clique */

  for ( j = 0; j < 8; j++ )
     for ( i = 0; i < 7; i++ )
        case_clic[i][j] = GTK_SHADOW_OUT;

  select_semantic = -1;

  /* Pas de slection dans le clipboard */

  select_reset();

  for ( i = 0; i < 31; i++ )
    {
      calendrier->clip_board.dates_copy[i] = g_date_new();
      calendrier->clip_board.dates_paste[i] = g_date_new();
    }

  calendrier->clip_board.nb_jours_copy = 0;
  calendrier->clip_board.nb_jours_paste = 0;
  calendrier->clip_board.semantique = PAS_DEFINIE;

  clic_i = clic_j = 0;

  date_tampon = g_date_new();

  /* Creation of the context_menu widget. The "show_all" of the menu
     is mandatory since it has no parent and hence, can't inherit it.*/

  context_menu = GTK_MENU( gtk_menu_new() );

  item_copy  = GTK_MENU_ITEM( gtk_menu_item_new_with_label("copy") );
  item_paste = GTK_MENU_ITEM( gtk_menu_item_new_with_label("paste") );

  gtk_menu_append( context_menu, GTK_WIDGET(item_copy) );
  gtk_menu_append( context_menu, GTK_WIDGET(item_paste) );

  gtk_signal_connect( GTK_OBJECT(item_copy), "activate",
                      GTK_SIGNAL_FUNC(copy_call_back),
                      NULL );

  gtk_signal_connect( GTK_OBJECT(item_paste), "activate",
                      GTK_SIGNAL_FUNC(paste_call_back),
                      NULL );

  gtk_widget_show_all( GTK_WIDGET(context_menu) );
}

/*-------------------------------------------------------------------------*/

static void gtk_calendrier_destroy( GtkObject *the_object )
{
  GtkCalendrier *calendrier;

  g_return_if_fail( the_object != NULL );
  g_return_if_fail( GTK_IS_CALENDRIER(the_object) );

  calendrier = GTK_CALENDRIER(the_object);

  /* Freeing the ressources used by the widget */

  for ( i = 0; i < 31; i++ )
    {
      g_date_free( calendrier->clip_board.dates_copy[i] );
      g_date_free( calendrier->clip_board.dates_paste[i] );
    }

  g_date_free( date_tampon );

  /* Chaining up for the parents */

  if ( GTK_OBJECT_CLASS(parent_class)->destroy )
    (* GTK_OBJECT_CLASS(parent_class)->destroy)(the_object);
}

/*-------------------------------------------------------------------------*/

GtkWidget * gtk_calendrier_new( void )
{
  GtkWidget *calendrier;
  calendrier = gtk_type_new( gtk_calendrier_get_type() );

  return GTK_WIDGET( calendrier );
}

/*-------------------------------------------------------------------------*/

static void gtk_calendrier_realize( GtkWidget *widget )
{
  GtkCalendrier *calendrier;
  GdkWindowAttr attributes;
  gint attributes_mask;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_CALENDRIER (widget));

  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

  calendrier = GTK_CALENDRIER (widget);

  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.event_mask = gtk_widget_get_events( widget ) | GDK_EXPOSURE_MASK |
                          GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
                          GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK;

  attributes.visual = gtk_widget_get_visual( widget );
  attributes.colormap = gtk_widget_get_colormap( widget );

  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
  widget->window = gdk_window_new( widget->parent->window, &attributes, attributes_mask );

  widget->style = gtk_style_attach( widget->style, widget->window );

  gdk_window_set_user_data( widget->window, widget );

  gtk_style_set_background( widget->style, widget->window, GTK_STATE_ACTIVE );
}

/*-------------------------------------------------------------------------*/

static void gtk_calendrier_size_request(GtkWidget *widget, GtkRequisition *requisition)
{
  requisition->width = ( 14 + 2 * gdk_char_width( widget->style->font, '0') )
                      * 7 + 1;
  requisition->height = ( 12 + gdk_char_height( widget->style->font, '0') )
                      * 9 + 1;
  case_width = requisition->width / 7;
  case_height = requisition->height / 9;
}

/*-------------------------------------------------------------------------*/

static void gtk_calendrier_size_allocate(GtkWidget *widget, GtkAllocation *allocation)
{
  GtkCalendrier *calendrier;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_CALENDRIER (widget));
  g_return_if_fail (allocation != NULL);

  widget->allocation = *allocation;

  if ( GTK_WIDGET_REALIZED(widget) )
    {
      calendrier = GTK_CALENDRIER (widget);

      gdk_window_move_resize( widget->window,
                              allocation->x, allocation->y,
                              allocation->width, allocation->height );

      case_width = widget->allocation.width / 7;
      case_height = widget->allocation.height / 9;
    }
}

/*-------------------------------------------------------------------------*/

static gint gtk_calendrier_expose(GtkWidget *widget, GdkEventExpose *event)
{
  GtkCalendrier *calendrier;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_CALENDRIER (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  calendrier = GTK_CALENDRIER (widget);

  gtk_calendrier_paint( widget, &rectangle );

  return FALSE;
}

/*-------------------------------------------------------------------------*/

static void gtk_calendrier_draw( GtkWidget *widget, GdkRectangle *area )
{
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_CALENDRIER (widget));
  g_return_if_fail (area != NULL);

  gtk_calendrier_paint( widget, area );
}

/*-------------------------------------------------------------------------*/

static void gtk_calendrier_paint( GtkWidget *widget, GdkRectangle *area )
{
  GtkCalendrier *calendrier;
  gint  x, y, long_txt, haut_txt;
  gchar tampon[5];
  GdkPoint fleche[3];

  g_return_if_fail (widget != NULL);
  g_return_if_fail (widget->window != NULL);
  g_return_if_fail (GTK_IS_CALENDRIER (widget));

  calendrier = GTK_CALENDRIER (widget);

  the_pixmap = gdk_pixmap_new( widget->window,
                               widget->allocation.width,
                               widget->allocation.height, -1 );

  /* We use a private gc based on the widget mid one for the forced colors */

  the_gc = gdk_gc_new( the_pixmap );
  gdk_gc_copy( the_gc, widget->style->mid_gc[widget->state] );

  /* Draw background of the whole widget surface */

  gdk_draw_rectangle( the_pixmap,
                      widget->style->mid_gc[widget->state],
                      TRUE,
                      0, 0,
                      widget->allocation.width,
                      widget->allocation.height );

  /* Dessin des jours de la matrice calcule. La hauteur de fonte est ici
    calcule sur la base d'un chiffre pour une meilleure adaptation. */

  haut_txt = gdk_char_height( widget->style->font, '0');
  k = 0;
  for ( j = 0; j < 6; j++ )
     for ( i = 0; i < 7; i++ )
       {
         /* Fond de la "case" gris ou blanc selon selection ou non. Comme
         le fond de la widget est dj gris, il n'y a rien  faire si la
         case appartient  la zone slectionne du clipboard. */

         x = i * case_width;
         y = ( j + 3 ) * case_height;

         if ( matrix_selected[i][j] == 0 ) /* Not selected : white */
          gdk_draw_rectangle( the_pixmap,
                              widget->style->white_gc,
                              TRUE,
                              x, y, case_width, case_height );

         /* Calcul longueur et hauteur du texte pour le centrage. */

         long_txt = gdk_string_measure( widget->style->font,
                                        calendrier->jours_mois[k] );
         x = x + ( case_width - long_txt ) / 2;

         calendrier->hauteur_texte = gdk_string_height( widget->style->font,
                                                        calendrier->jours_mois[k] );
         y = y + ( case_height - haut_txt ) / 2 + haut_txt;

         if ( strcmp(calendrier->jours_mois[k], "0") ) /* Si "0", on dessine rien */
           {
             /* Le jour courant est marqu par une fonte grasse pour qu'on
                s'y retrouve dans le programmateur. Plutt que de charger
                une fonte spciale, on conserve la fonte par dfaut, en
                dessinant le jour deux fois en dcalant de 1 pixel  gauche:
                on obtient ainsi du "bold" simple et pour pas chr...
             */

             if ( atoi(calendrier->jours_mois[k]) ==
                  g_date_day(calendrier->date_interne) )

                gdk_draw_string( the_pixmap, widget->style->font,
                                 widget->style->fg_gc[widget->state],
                                 x - 1, y, calendrier->jours_mois[k] );

             gdk_draw_string( the_pixmap, widget->style->font,
                              widget->style->fg_gc[widget->state],
                              x, y, calendrier->jours_mois[k] );
           }
         k++;
       }

  /* Dessin du quadrillage et effet 3D de la matrice des jours */

  for ( j = 4 * case_height; j < 9 * case_height; j += case_height )

     gdk_draw_line( the_pixmap, widget->style->mid_gc[widget->state],
                    0 , j, 7 * case_width, j );

  for ( i = case_width; i < 7 * case_width; i += case_width )

     gdk_draw_line( the_pixmap, widget->style->mid_gc[widget->state],
                    i , 3 * case_height, i, 9 * case_height );

  gtk_draw_shadow( widget->style, the_pixmap, widget->state,
                   GTK_SHADOW_IN,
		               0, 3 * case_height,
                   7 * case_width, 6 * case_height );

  /* Dessin du header : Couleur de fond */

  gdk_draw_rectangle( the_pixmap,
                      widget->style->mid_gc[widget->state],
                      TRUE,
                      0, 0,
                      7 * case_width, 3 * case_height );

  /* Dessin du header : "Boutons" des jours de la semaine */

  for ( i = 0; i < 7; i++ )
     {
       gtk_draw_shadow( widget->style, the_pixmap, widget->state,
                        case_clic[i][2], /* Shadow IN-OUT */
                        i * case_width, 2 * case_height,
                        case_width, case_height );

       long_txt = gdk_string_measure( widget->style->font, calendrier->jours_semaine[i] );
       x = i * case_width + ( case_width - long_txt ) / 2;
       y = 2 * case_height + ( case_height - calendrier->hauteur_texte ) / 2 + calendrier->hauteur_texte;

       gdk_draw_string( the_pixmap, widget->style->font,
                        widget->style->fg_gc[widget->state],
                        x, y,
                        calendrier->jours_semaine[i] );
     }

  /* Dessin du header : "Boutons" semaine et week-end */

  gtk_draw_shadow( widget->style, the_pixmap, widget->state,
                   case_clic[0][1], /* Shadow IN-OUT */
                   0, case_height,
                   5 * case_width, case_height );

  /* Calcul longueur et hauteur du texte pour le centrage. */

  long_txt = gdk_string_measure( widget->style->font, "Semaine" );
  x = ( 5 * case_width - long_txt ) / 2;
  y = case_height + ( case_height - calendrier->hauteur_texte ) / 2
    + calendrier->hauteur_texte;

  gdk_draw_string( the_pixmap, widget->style->font,
                        widget->style->fg_gc[widget->state],
                        x, y, "Semaine" );

  gtk_draw_shadow( widget->style, the_pixmap, widget->state,
                   case_clic[6][1],
                   5 * case_width, case_height,
                   2 * case_width, case_height );

  /* Calcul longueur et hauteur du texte pour le centrage. */

  long_txt = gdk_string_measure( widget->style->font, "W.E." );
  x = 5 * case_width + ( 2 * case_width - long_txt ) / 2;
  y = case_height + ( case_height - calendrier->hauteur_texte ) / 2
    + calendrier->hauteur_texte;

  gdk_draw_string( the_pixmap, widget->style->font,
                        widget->style->fg_gc[widget->state],
                        x, y, "W.E." );

  /* Dessin du header : Slection du mois et affichage anne. */

  /*--- flche gauche ---*/

  gtk_draw_shadow( widget->style, the_pixmap, widget->state,
                   /*GTK_SHADOW_OUT,*/
                   case_clic[0][0],
                   0, 0, case_width, case_height );

  fleche[0].x = case_width / 4;
  fleche[0].y = case_height / 2;
  fleche[1].x = 5 * case_width / 8;
  fleche[1].y = case_height / 4;
  fleche[2].x = 5 * case_width / 8;
  fleche[2].y = 3 * case_height / 4;

  gdk_draw_polygon( the_pixmap,
                    widget->style->fg_gc[widget->state],
                    TRUE, fleche, 3 );


  /*--- mois date interne ---*/

  gdk_draw_rectangle( the_pixmap, widget->style->white_gc,
                      TRUE,
                      case_width, 0, 3 * case_width, case_height );

  gtk_draw_shadow( widget->style, the_pixmap, widget->state,
                   GTK_SHADOW_IN,
                   case_width, 0, 3 * case_width, case_height );

  /* Calcul longueur et hauteur du texte pour le centrage. */

  long_txt = gdk_string_measure( widget->style->font,
                                 calendrier->mois_annee[g_date_month(calendrier->date_interne)-1] );

  x = case_width + ( 3 * case_width - long_txt ) / 2;
  y = ( case_height - calendrier->hauteur_texte ) / 2
    + calendrier->hauteur_texte;

  gdk_draw_string( the_pixmap, widget->style->font,
                   widget->style->fg_gc[widget->state],
                   x, y,
                   calendrier->mois_annee[g_date_month(calendrier->date_interne) - 1] );

  /*--- flche droite ---*/

  gtk_draw_shadow( widget->style, the_pixmap, widget->state,
                   case_clic[4][0],
                   4 * case_width, 0, case_width, case_height );

  fleche[0].x = 4 * case_width + 3 * case_width / 8;
  fleche[0].y = case_height / 4;
  fleche[1].x = 4 * case_width + 3 * case_width / 4;
  fleche[1].y = case_height / 2;
  fleche[2].x = 4 * case_width + 3 * case_width / 8;
  fleche[2].y = 3 * case_height / 4;

  gdk_draw_polygon( the_pixmap,
                    widget->style->fg_gc[widget->state],
                    TRUE, fleche, 3 );

  /*--- anne date interne ---*/

  gdk_draw_rectangle( the_pixmap, widget->style->white_gc, TRUE,
                      5 * case_width, 0,
                      2 * case_width, case_height );

  gtk_draw_shadow( widget->style, the_pixmap, widget->state,
                   GTK_SHADOW_IN,
                   5 * case_width, 0,
                   2 * case_width, case_height );

  sprintf(tampon, "%d", g_date_year(calendrier->date_interne) );

  /* Calcul longueur et hauteur du texte pour le centrage. */

  long_txt = gdk_string_measure( widget->style->font, tampon );
  x = 5 * case_width + ( 2 * case_width - long_txt ) / 2;
  y = ( case_height - calendrier->hauteur_texte ) / 2
    + calendrier->hauteur_texte;

  gdk_draw_string( the_pixmap, widget->style->font,
                   widget->style->fg_gc[widget->state],
                   x, y, tampon );

  /* Blitting du pixmap  la fentre pour un affichage instantann. */

  gdk_draw_pixmap( widget->window,
                   widget->style->fg_gc[widget->state],
                   the_pixmap,
                   0, 0, 0, 0, -1, -1 );
  gdk_gc_unref(the_gc);
  gdk_pixmap_unref(the_pixmap);
}

/*-------------------------------------------------------------------------*/

static gint gtk_calendrier_button_press( GtkWidget *widget,
                                         GdkEventButton *event )
{
  GtkCalendrier *calendrier;

  gint mois;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_CALENDRIER (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  calendrier = GTK_CALENDRIER(widget);

  /* Identification de la case i,j clique */

  clic_i = div(event->x, case_width).quot;
  clic_j = div(event->y, case_height).quot;

  if ( clic_i > 6 ) clic_i = 6; /* C'est possible car la fentre est ... */
  if ( clic_j > 8 ) clic_j = 8; /* ... dimensionne par GTK+ */

  /* Prparation des zones graphiques multiples */

  if ( event->button == 1 ) /* left button: selections & clics */
    {

      switch( clic_j )
      {
        case 0: /* First row, where the month arrows are */
          case_clic[clic_i][clic_j] = GTK_SHADOW_IN;
          if ( clic_i == 0 ) /* flche gauche */
           {
             mois = g_date_month( calendrier->date_interne );
             if ( mois > 1 )
              {
                mois--;
                g_date_subtract_months( calendrier->date_interne, 1 );

                /* Si on se trouve sur le mois courant, on positionne la
                   date courante, sinon on positionne le 1er du mois. */

                if ( g_date_month(calendrier->date_interne) ==
                     g_date_month(calendrier->date_courante) )
                   g_date_set_day( calendrier->date_interne,
                                   g_date_day(calendrier->date_courante) );

                else g_date_set_day( calendrier->date_interne, 1 );

                calcule_date_matrice( calendrier->jours_mois,
                                      calendrier->date_interne );

                /* new month: selection is cleared and clipboard is
                   transposed into a new selection if not empty */

                select_reset(); /* It will be a new selection... */

                if ( calendrier->clip_board.nb_jours_copy != 0 )
                  select_transpose();
                else select_semantic = PAS_DEFINIE;

                gtk_signal_emit( GTK_OBJECT(widget),
                                 gtk_calendrier_signals[NEW_DATE_SELECTED] );
              }
           }
          else if ( clic_i == 4 ) /* flche droite */
                {
                  mois = g_date_month( calendrier->date_interne );
                  if ( mois < 12 )
                   {
                     mois++;
                     g_date_add_months( calendrier->date_interne, 1 );

                     /* Si on se trouve sur le mois courant, on positionne la
                        date courante, sinon on positionne le 1er du mois. */

                     if ( g_date_month(calendrier->date_interne) ==
                          g_date_month(calendrier->date_courante) )
                       g_date_set_day( calendrier->date_interne,
                                       g_date_day(calendrier->date_courante) );

                     else g_date_set_day( calendrier->date_interne, 1 );

                     calcule_date_matrice( calendrier->jours_mois,
                                           calendrier->date_interne );

                     select_reset();

                     if ( calendrier->clip_board.nb_jours_copy != 0 )
                       select_transpose();
                     else select_semantic = PAS_DEFINIE;

                     gtk_signal_emit( GTK_OBJECT(widget),
                                      gtk_calendrier_signals[NEW_DATE_SELECTED] );
                   }
                }
        break;

        case 1: /* buttons "semaine" or "WE" */

          select_reset(); /* Not mandatory:see the doc */

          if ( clic_i < 5 ) /* semaine */
           {
             if ( select_semantic != LA_SEMAINE )
               { /* Here, we change the semantic of the clipboard : so
                    all the previous selections are cancelled */
                 select_semantic = LA_SEMAINE;
                 calendrier->clip_board.nb_jours_copy = 0;
                 calendrier->clip_board.nb_jours_paste = 0;
                 calendrier->clip_board.semantique = PAS_DEFINIE;
               }

             k = 0; /* will be the 1st valid day encountered. We always check
                       the selected day before. If the action brings a whole
                       unselection, then no change of current day occurs.
                       But if only one day is selected, then it will be the
                       current day. */

             /* If the old selection is already in the clipboard we perform a
                new selection for the eventual paste.
                If not, the new selection will complete the old one.*/

             case_clic[0][1] = case_clic[1][1] = case_clic[2][1] =
             case_clic[3][1] = case_clic[4][1] = GTK_SHADOW_IN;

             /* Selection-unselection of the whole week days */

             for ( j = 0; j < 6; j++ )
               for ( i = 0; i < 5; i++ )
                 if ( matrix_selected[i][j] == 1 ) /* unselection */
                    matrix_selected[i][j] = 0;
                 else /* selection (if the day exists) and setting as current */
                 if ( strcmp(calendrier->jours_mois[i+7*j], "0") )
                  {
                    matrix_selected[i][j] = 1;
                    if ( k == 0 ) /* 1st day of the selection */
                      k = atoi( calendrier->jours_mois[i+7*j] );
                  }
           }
          else { /* WEEK END */

                 if ( select_semantic != LE_WEEK_END )
                   {
                     select_semantic = LE_WEEK_END;
                     calendrier->clip_board.nb_jours_copy = 0;
                     calendrier->clip_board.nb_jours_paste = 0;
                     calendrier->clip_board.semantique = PAS_DEFINIE;
                   }

                 k = 0;

                 case_clic[5][1] = case_clic[6][1] = GTK_SHADOW_IN;

                 /* Selection-unselection of the whole week-end days */

                 for ( j = 0; j < 6; j++ )
                   for ( i = 5; i < 7; i++ )
                     if ( matrix_selected[i][j] == 1 ) /* unselection */
                       matrix_selected[i][j] = 0;
                     else /* selection (if the day exists) and setting as current */
                     if ( strcmp( calendrier->jours_mois[i+7*j], "0" ) )
                      {
                        matrix_selected[i][j] = 1;
                        if ( k == 0 )
                           k = atoi( calendrier->jours_mois[i+7*j] );
                      }
               }

          /* As several days are selected, set the date to the 1st valid day encountered */

          if ( k != 0 )
            {
              g_date_set_day( calendrier->date_interne, k );

              gtk_signal_emit( GTK_OBJECT(widget),
                               gtk_calendrier_signals[NEW_DATE_SELECTED] );
            }
        break;

        case 2: /* Days of the week buttons */

          select_reset(); /* Not mandatory:see the doc */

          case_clic[clic_i][clic_j] = GTK_SHADOW_IN;

          if ( select_semantic != JOURS_SEMAINE )
           {
             calendrier->clip_board.nb_jours_copy = 0;
             calendrier->clip_board.nb_jours_paste = 0;
             calendrier->clip_board.semantique = PAS_DEFINIE;

             select_semantic = JOURS_SEMAINE;
           }

          k = 0;

          for ( j = 0; j < 6; j++ )
            if ( matrix_selected[clic_i][j] == 1 ) /* unselection */
               matrix_selected[clic_i][j] = 0;
            else /* selection (if the day exists) and setting as current */
            if ( strcmp(calendrier->jours_mois[clic_i+7*j], "0") )
             {
               matrix_selected[clic_i][j] = 1;
               if ( k == 0 )
                  k = atoi( calendrier->jours_mois[clic_i+7*j] );
             }

          if ( k != 0 )
            {
              g_date_set_day( calendrier->date_interne, k );

              gtk_signal_emit( GTK_OBJECT(widget),
                               gtk_calendrier_signals[NEW_DATE_SELECTED] );
            }

        break;

        default: /* Days of the month matrix */

          select_reset(); /* Not mandatory:see the doc */

          if ( select_semantic != JOURS_MOIS )
           {
             calendrier->clip_board.nb_jours_copy = 0;
             calendrier->clip_board.nb_jours_paste = 0;
             calendrier->clip_board.semantique = PAS_DEFINIE;

             select_semantic = JOURS_MOIS;
           }

          if ( matrix_selected[clic_i][clic_j-3] == 1 ) /* unselection */
           {
             matrix_selected[clic_i][clic_j-3] = 0;
           }

          else /* selection (if the day exists) and setting as current */
           {
             if ( strcmp(calendrier->jours_mois[clic_i+7*(clic_j-3)], "0") )
               {
                 matrix_selected[clic_i][clic_j-3] = 1;
                 g_date_set_day( calendrier->date_interne,
                                 atoi( calendrier->jours_mois[clic_i+7*(clic_j-3)] ) );
                 gtk_signal_emit( GTK_OBJECT(widget),
                                  gtk_calendrier_signals[NEW_DATE_SELECTED] );
               }
           }
        break;
      }

      gtk_calendrier_paint( GTK_WIDGET(calendrier), &rectangle );

    } /* End if left button */

  else /* Other button (right): context menu. Threatment on callbacks */

    gtk_menu_popup( context_menu,
                    NULL, NULL, NULL, NULL,
                    event->button,
                    event->time );

  return FALSE;
}

/*-------------------------------------------------------------------------*/

static gint gtk_calendrier_button_release( GtkWidget *widget,
                                           GdkEventButton *event )
{
  GtkCalendrier *calendrier;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_CALENDRIER (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  calendrier = GTK_CALENDRIER(widget);

  case_clic[clic_i][clic_j] = GTK_SHADOW_OUT;

  /* Release ventuelle des boutons multiples */

  if ( clic_j == 1 )
    { if ( clic_i < 5 )
           case_clic[0][1] =
           case_clic[1][1] =
           case_clic[2][1] =
           case_clic[3][1] =
           case_clic[4][1] = GTK_SHADOW_OUT;

      else case_clic[5][1] =
           case_clic[6][1] = GTK_SHADOW_OUT;
    }

  /* Release soit d'un bouton d'une case, soit d'une slection de jour */

  case_clic[clic_i][clic_j] = GTK_SHADOW_OUT;

  gtk_calendrier_paint( GTK_WIDGET(calendrier), &rectangle );

  return FALSE;
}

/*-------------------------------------------------------------------------*/

gint gtk_calendrier_get_clipboard_copy( GtkCalendrier *calendrier,
                                               GDate *the_dates[] )
{
  g_return_val_if_fail( calendrier != NULL, -1 );
  g_return_val_if_fail( the_dates != NULL, -1 );

  for( i = 0; i < calendrier->clip_board.nb_jours_copy; i++ )
     the_dates[i] = calendrier->clip_board.dates_copy[i];

  return calendrier->clip_board.nb_jours_copy;
}

/*-------------------------------------------------------------------------*/

gint gtk_calendrier_get_clipboard_paste( GtkCalendrier *calendrier,
                                                 GDate *the_dates[] )
{
  g_return_val_if_fail( calendrier != NULL, -1 );
  g_return_val_if_fail( the_dates != NULL, -1 );

  for( i = 0; i < calendrier->clip_board.nb_jours_paste; i++ )
     the_dates[i] = calendrier->clip_board.dates_paste[i];

  return calendrier->clip_board.nb_jours_paste;
}

/*-------------------------------------------------------------------------*/

gint gtk_calendrier_clear_clipboard( GtkCalendrier *calendrier )
{
  g_return_val_if_fail( calendrier != NULL, -1 );

  calendrier->clip_board.nb_jours_copy = 0;
  calendrier->clip_board.nb_jours_paste = 0;
  calendrier->clip_board.semantique = PAS_DEFINIE;

  return 0;
}

/*-------------------------------------------------------------------------*/

gint gtk_calendrier_get_internal_date( GtkCalendrier *calendrier,
                                               GDate *the_date )
{
  g_return_val_if_fail( calendrier != NULL, -1 );
  g_return_val_if_fail( the_date != NULL, -1 );

  /* The internal date is always valid : maintained as such */

  g_date_set_julian( the_date, g_date_julian( calendrier->date_interne ) );

  return 0;
}

/*-------------------------------------------------------------------------*/

gint gtk_calendrier_set_internal_date( GtkCalendrier *calendrier,
                                               GDate *the_date )
{
  g_return_val_if_fail( calendrier != NULL, -1 );
  g_return_val_if_fail( the_date != NULL, -1 );
  g_return_val_if_fail( g_date_valid( the_date ), -1 );

  g_date_set_julian( calendrier->date_interne, g_date_julian( the_date ) );
  calcule_date_matrice( calendrier->jours_mois, calendrier->date_interne );

  gtk_calendrier_paint( GTK_WIDGET(calendrier), &rectangle );

  return 0;
}

/*-------------------------------------------------------------------------*/

gint gtk_calendrier_get_clipboard_semantic( GtkCalendrier *calendrier )
{
  g_return_val_if_fail( calendrier != NULL, -1 );

  return calendrier->clip_board.semantique;
}

/*-------------------------------------------------------------------------*/

gint gtk_calendrier_get_days_selected( GtkCalendrier *calendrier,
                                               guint the_days[] )
{
  g_return_val_if_fail( calendrier != NULL, -1 );
  g_return_val_if_fail( the_days != NULL, -1 );

  k = 0;

  for ( j = 0; j < 6; j++ )
    for ( i = 0; i < 7; i++ )
      if ( matrix_selected[i][j] != 0 )
        {
          the_days[k++] = atoi( calendrier->jours_mois[i+7*j] );
        }
  return k;
}
