/***************************************************************************
                          gtk_flag_binaire.c  -  description
                             -------------------
    begin                : Sun Jan 28 2001
    copyright            : (C) 2001 by Dominique Laigle
    email                : lainal@gmx.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "gtkflagbinaire.h"

/* Private data and functions invisible from the interface point of view.
   That's why they are static too. This way of defining private data is
   probably not the preconized one by GTK developers, but I find it clear.
 */

/* For drawing facilitites */

static GdkRectangle rectangle;
static GdkGC       *the_gc;
static GdkColormap *the_colormap;

/* Privates common datas */

static gint  buf_nb_flags,
             larg_cell, haut_flag, haut_txt,
             idx_flag;
static gchar **buf_labels;

/*-------- Forward References ----------------------------------*/

static GtkWidgetClass *parent_class = NULL;

static void gtk_flag_binaire_class_init(GtkFlagBinaireClass *);
static void gtk_flag_binaire_init(GtkFlagBinaire *);
static void gtk_flag_binaire_realize(GtkWidget *);
static void gtk_flag_binaire_unrealize(GtkWidget *);
static void gtk_flag_binaire_destroy( GtkObject * );

static void gtk_flag_binaire_size_request(GtkWidget *, GtkRequisition *);
static void gtk_flag_binaire_size_allocate(GtkWidget *, GtkAllocation *);

static void gtk_flag_binaire_paint(GtkWidget *, GdkRectangle *);
static void gtk_flag_binaire_draw(GtkWidget *, GdkRectangle *);
static gint gtk_flag_binaire_expose(GtkWidget *, GdkEventExpose *);

/*------------------------------------------------------------------------*/

GtkType gtk_flag_binaire_get_type( void )
{
  static GtkType flag_binaire_type = 0;

  if ( !flag_binaire_type )
    {
      static const GtkTypeInfo flag_binaire_info =
      {
	      "GtkFlagBinaire",
	      sizeof (GtkFlagBinaire),
	      sizeof (GtkFlagBinaireClass),
	      (GtkClassInitFunc)  gtk_flag_binaire_class_init,
	      (GtkObjectInitFunc) gtk_flag_binaire_init,
	      (GtkArgSetFunc) NULL,
	      (GtkArgGetFunc) NULL,
      };
      flag_binaire_type = gtk_type_unique( GTK_TYPE_WIDGET, &flag_binaire_info );
    }
  return flag_binaire_type;
}

/*-------------------------------------------------------------------------*/

static void gtk_flag_binaire_class_init( GtkFlagBinaireClass *klass )
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;

  object_class = (GtkObjectClass*) klass;
  widget_class = (GtkWidgetClass*) klass;
  parent_class = gtk_type_class( GTK_TYPE_WIDGET );

  widget_class->realize   = gtk_flag_binaire_realize;
  widget_class->unrealize = gtk_flag_binaire_unrealize;
  object_class->destroy   = gtk_flag_binaire_destroy;

  widget_class->draw         = gtk_flag_binaire_draw;
  widget_class->expose_event = gtk_flag_binaire_expose;

  widget_class->size_request  = gtk_flag_binaire_size_request;
  widget_class->size_allocate = gtk_flag_binaire_size_allocate;

  widget_class->button_press_event   = NULL;
  widget_class->button_release_event = NULL;
  widget_class->motion_notify_event  = NULL;
}

/*-------------------------------------------------------------------------*/

static void gtk_flag_binaire_init( GtkFlagBinaire *the_flag_binaire )
{
  GtkWidget *the_widget;

  the_widget = GTK_WIDGET(the_flag_binaire);
  GTK_WIDGET_SET_FLAGS (the_widget, GTK_CAN_FOCUS);

  /* Initializations of the data of the widget */

  the_flag_binaire->nb_flags = buf_nb_flags;

  the_colormap = gdk_colormap_get_system(); /* For the change of colors */

  /* Default ON-OFF colors */

  gdk_color_parse( "black", &the_flag_binaire->off_color );
  gdk_color_parse( "darkRed", &the_flag_binaire->on_color );

  /* Memory allocations regarding the nb of flags given at the creation. */

  the_flag_binaire->indicator = g_new( flag_t, buf_nb_flags );

  /* A bit of initializations */

  for ( idx_flag = 0; idx_flag < buf_nb_flags; idx_flag++ )
    {
      the_flag_binaire->indicator[idx_flag].label = g_new( gchar, LABEL_MAX_LENGTH );
      g_snprintf( the_flag_binaire->indicator[idx_flag].label, 16,
                  "%s",buf_labels[idx_flag] );
      the_flag_binaire->indicator[idx_flag].flag = FALSE;
    }
}

/*-------------------------------------------------------------------------*/

static void gtk_flag_binaire_destroy( GtkObject *the_object )
{
  GtkFlagBinaire *the_flag_binaire;

  g_return_if_fail( the_object != NULL );
  g_return_if_fail( GTK_IS_FLAG_BINAIRE(the_object) );

  the_flag_binaire = GTK_FLAG_BINAIRE(the_object);

  /* Memory deallocations */

  for ( idx_flag = 0; idx_flag < buf_nb_flags; idx_flag++ )
      g_free( the_flag_binaire->indicator[idx_flag].label );

  g_free( the_flag_binaire->indicator );

  /* Chaining up for the parents */

  if ( GTK_OBJECT_CLASS(parent_class)->destroy )
    (* GTK_OBJECT_CLASS(parent_class)->destroy)(the_object);
}

/*-------------------------------------------------------------------------*/

GtkWidget * gtk_flag_binaire_new( guint nb_flags, gchar **labels )
{
  GtkWidget *the_flag_binaire;

  g_return_val_if_fail( nb_flags != 0, NULL ); /* Ha, ha, ha ! */
  g_return_val_if_fail( nb_flags <= MAX_FLAGS_ALLOWED, NULL );

  /* Store the "constructor" arg. to use it at the init step. Got to be
  assigned before the instance exists...*/

  buf_nb_flags = nb_flags;
  buf_labels = labels;

  the_flag_binaire = gtk_type_new( gtk_flag_binaire_get_type() );

  return GTK_WIDGET( the_flag_binaire );
}

/*-------------------------------------------------------------------------*/

static void gtk_flag_binaire_realize( GtkWidget *the_widget )
{
  GtkFlagBinaire *the_flag_binaire;
  GdkWindowAttr attributes;
  gint attributes_mask;

  g_return_if_fail( the_widget != NULL );
  g_return_if_fail( GTK_IS_FLAG_BINAIRE( the_widget ) );

  GTK_WIDGET_SET_FLAGS( the_widget, GTK_REALIZED );

  the_flag_binaire = GTK_FLAG_BINAIRE( the_widget );

  attributes.x = the_widget->allocation.x;
  attributes.y = the_widget->allocation.y;
  attributes.width = the_widget->allocation.width;
  attributes.height = the_widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.event_mask = gtk_widget_get_events( the_widget ) | GDK_EXPOSURE_MASK;

  attributes.visual = gtk_widget_get_visual( the_widget );
  attributes.colormap = gtk_widget_get_colormap( the_widget );

  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
  the_widget->window = gdk_window_new( the_widget->parent->window, &attributes, attributes_mask );

  the_widget->style = gtk_style_attach( the_widget->style, the_widget->window );

  gdk_window_set_user_data( the_widget->window, the_widget );

  gtk_style_set_background( the_widget->style, the_widget->window, GTK_STATE_ACTIVE );
}

/*-------------------------------------------------------------------------*/

static void gtk_flag_binaire_unrealize (GtkWidget *the_widget)
{
  GtkFlagBinaire *the_flag_binaire;

  g_return_if_fail( the_widget != NULL );
  g_return_if_fail( GTK_IS_FLAG_BINAIRE(the_widget) );

  the_flag_binaire = GTK_FLAG_BINAIRE(the_widget);

  if ( GTK_WIDGET_MAPPED(the_widget) )
     gtk_widget_unmap(the_widget);

  GTK_WIDGET_UNSET_FLAGS( the_widget, GTK_MAPPED );

  /* destroy the widget->window and unset the realized flag */

  if ( GTK_WIDGET_CLASS(parent_class)->unrealize )
    (* GTK_WIDGET_CLASS(parent_class)->unrealize)( the_widget );
}

/*-------------------------------------------------------------------------*/

static void gtk_flag_binaire_size_request(GtkWidget *the_widget, GtkRequisition *requisition)
{
  /* Sizing policy is :
    - legend height related to 2 times the height of the used police
    - flag height constat to 1 time a char height of the used police
  */

  larg_cell = 6 + gdk_string_width( the_widget->style->font, "label of 16 char");

  requisition->width = buf_nb_flags * larg_cell;

  haut_txt = 2 * ( gdk_string_height( the_widget->style->font, "g4" ) );
  haut_flag = gdk_char_height( the_widget->style->font, 'O' );

  requisition->height = haut_txt + haut_flag;
}

/*-------------------------------------------------------------------------*/

static void gtk_flag_binaire_size_allocate(GtkWidget *the_widget, GtkAllocation *allocation)
{
  GtkFlagBinaire *the_flag_binaire;

  g_return_if_fail (the_widget != NULL);
  g_return_if_fail (GTK_IS_FLAG_BINAIRE (the_widget));
  g_return_if_fail (allocation != NULL);

  the_widget->allocation = *allocation;

  if (GTK_WIDGET_REALIZED (the_widget))
    {
      the_flag_binaire = GTK_FLAG_BINAIRE(the_widget);

      /* Resize policy is that the flag heights stay constant as
         they were requested. The space left is allocated to the title.
      */
      larg_cell = allocation->width / buf_nb_flags;

      haut_txt = allocation->height - haut_flag;

      gdk_window_move_resize( the_widget->window,
                              allocation->x, allocation->y,
                              allocation->width,
                              allocation->height );
    }
}

/*-------------------------------------------------------------------------*/

static gint gtk_flag_binaire_expose(GtkWidget *the_widget, GdkEventExpose *event)
{
  GtkFlagBinaire *the_flag_binaire;

  g_return_val_if_fail (the_widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_FLAG_BINAIRE (the_widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  the_flag_binaire = GTK_FLAG_BINAIRE (the_widget);

  /* Recuperation de la taille de la fenetre */

  rectangle.x = 0;
  rectangle.y = 0;
  rectangle.width = the_widget->allocation.width;
  rectangle.height = the_widget->allocation.height;

  /* Appel de la fonction de dessin proprement dit */

  gtk_flag_binaire_paint( the_widget, &rectangle );

  return FALSE;
}

/*-------------------------------------------------------------------------*/

static void gtk_flag_binaire_draw( GtkWidget *the_widget, GdkRectangle *area )
{
  g_return_if_fail (the_widget != NULL);
  g_return_if_fail (GTK_IS_FLAG_BINAIRE (the_widget));
  g_return_if_fail (area != NULL);

  gtk_flag_binaire_paint( the_widget, area );
}

/*-------------------------------------------------------------------------*/

static void gtk_flag_binaire_paint( GtkWidget *the_widget, GdkRectangle *area )
{
  GtkFlagBinaire *the_flag_binaire;

  g_return_if_fail (the_widget != NULL);
  g_return_if_fail (the_widget->window != NULL);
  g_return_if_fail (GTK_IS_FLAG_BINAIRE (the_widget));

  the_flag_binaire = GTK_FLAG_BINAIRE(the_widget);

  /*
    Since this widget is not an interactive one, we don't use any pixmap for
    the drawing : that way we don't waste resources from X server.
  */

  /* we use a private gc based on the widget mid one for the forced colors. */

  the_gc = gdk_gc_new( the_widget->window );
  gdk_gc_copy( the_gc, the_widget->style->mid_gc[the_widget->state] );

  /* Background of the widget with the default color */

  gdk_draw_rectangle( the_widget->window, the_gc, TRUE,
                      0, 0,
                      the_widget->allocation.width,
                      the_widget->allocation.height );

  /* Draw the cycle states and labels */

  for ( idx_flag = 0; idx_flag < buf_nb_flags; idx_flag++ )
    {
      if ( the_flag_binaire->indicator[idx_flag].flag == TRUE )
        {
         /* Construction and allocation of the "ON" color */

         gdk_colormap_alloc_color( the_colormap, &the_flag_binaire->on_color, FALSE, TRUE );

         gdk_gc_set_foreground( the_gc, &the_flag_binaire->on_color );

         gdk_draw_rectangle( the_widget->window, the_gc, TRUE,
                             idx_flag * larg_cell, haut_txt,
                             larg_cell, haut_flag );

         /* freeing the ON color previously allocated */

         gdk_colormap_free_colors( the_colormap, &the_flag_binaire->on_color, 1 );
        }

      else
        {
         /* Construction and allocation of the "OFF" color */

         gdk_colormap_alloc_color( the_colormap, &the_flag_binaire->off_color, FALSE, TRUE );

         gdk_gc_set_foreground( the_gc, &the_flag_binaire->off_color );

         gdk_draw_rectangle( the_widget->window, the_gc, TRUE,
                             idx_flag * larg_cell, haut_txt,
                             larg_cell, haut_flag );

         /* freeing the OFF color previously allocated */

         gdk_colormap_free_colors( the_colormap, &the_flag_binaire->off_color, 1 );
        }

      /* 3D effect for the "LED" */

      gtk_draw_shadow( the_widget->style, the_widget->window, the_widget->state,
                       GTK_SHADOW_IN,
                       idx_flag * larg_cell, haut_txt,
                       larg_cell, haut_flag );

      /* Drawing label without frame decoration */

      gdk_draw_string( the_widget->window, the_widget->style->font,
                       the_widget->style->fg_gc[the_widget->state],
                       idx_flag * larg_cell +
                       ( ( larg_cell -
                           gdk_string_width( the_widget->style->font,
                                             the_flag_binaire->indicator[idx_flag].label )
                         ) / 2 ),
                       ( haut_txt +
                         gdk_char_height( the_widget->style->font, 'n' )
                       ) / 2,
                       the_flag_binaire->indicator[idx_flag].label );
    }

  gdk_gc_unref(the_gc);
}

/*-------------------------------------------------------------------------*/

void gtk_flag_binaire_set_flag( GtkFlagBinaire *the_flag_binaire,
                                guint id_flag )
{
  g_return_if_fail( the_flag_binaire != NULL );
  g_return_if_fail( GTK_IS_FLAG_BINAIRE( the_flag_binaire ) );

  g_return_if_fail( id_flag < the_flag_binaire->nb_flags );

  the_flag_binaire->indicator[id_flag].flag = TRUE;

  if ( GTK_WIDGET_MAPPED(the_flag_binaire) )
    gtk_flag_binaire_paint( GTK_WIDGET(the_flag_binaire), &rectangle );
}

/*-------------------------------------------------------------------------*/

void gtk_flag_binaire_clear_flag( GtkFlagBinaire *the_flag_binaire,
                                  guint           id_flag )
{
  g_return_if_fail( the_flag_binaire != NULL );
  g_return_if_fail( GTK_IS_FLAG_BINAIRE( the_flag_binaire ) );

  g_return_if_fail( id_flag < the_flag_binaire->nb_flags );

  the_flag_binaire->indicator[id_flag].flag = FALSE;

  if ( GTK_WIDGET_MAPPED(the_flag_binaire) )
    gtk_flag_binaire_paint( GTK_WIDGET(the_flag_binaire), &rectangle );
}

/*-------------------------------------------------------------------------*/

void gtk_flag_binaire_set_label( GtkFlagBinaire *the_flag_binaire,
                                 guint id_flag,
                                 const gchar *new_label )
{
  g_return_if_fail( the_flag_binaire != NULL );
  g_return_if_fail( GTK_IS_FLAG_BINAIRE( the_flag_binaire ) );

  g_return_if_fail( id_flag < the_flag_binaire->nb_flags );

  g_snprintf( the_flag_binaire->indicator[id_flag].label, 16, "%s", new_label );

  /* Redraw the widget only if already mapped. The user could call this
     method before calling any gtk_widget_show()... */

  if ( GTK_WIDGET_MAPPED(the_flag_binaire) )
    gtk_flag_binaire_paint( GTK_WIDGET(the_flag_binaire), &rectangle );
}

/*-------------------------------------------------------------------------*/

void gtk_flag_binaire_set_on_color( GtkFlagBinaire *the_flag_binaire,
                                    const gchar *color_name )
{
  g_return_if_fail( the_flag_binaire != NULL );
  g_return_if_fail( GTK_IS_FLAG_BINAIRE( the_flag_binaire ) );

  /* will abort if the color_name is not defined in X'rgb.txt */

  gdk_color_parse( color_name, &the_flag_binaire->on_color );

  if ( GTK_WIDGET_MAPPED(the_flag_binaire) )
    gtk_flag_binaire_paint( GTK_WIDGET(the_flag_binaire), &rectangle );
}

/*-------------------------------------------------------------------------*/

void gtk_flag_binaire_set_off_color( GtkFlagBinaire *the_flag_binaire,
                                        const gchar *color_name )
{
  g_return_if_fail( the_flag_binaire != NULL );
  g_return_if_fail( GTK_IS_FLAG_BINAIRE( the_flag_binaire ) );

  /* will abort if the color_name is not defined in X'rgb.txt */

  gdk_color_parse( color_name, &the_flag_binaire->off_color );

  if ( GTK_WIDGET_MAPPED(the_flag_binaire) )
    gtk_flag_binaire_paint( GTK_WIDGET(the_flag_binaire), &rectangle );
}
