/***************************************************************************
                          gtk_programmateur.c  -  description
                             -------------------
    begin                : Sun Jan 28 2001
    copyright            : (C) 2001 by Dominique Laigle
    email                : lainal@gmx.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include "gtkprogrammateur.h"

#define FULL_DAY 96 /* 4 x 24 */

/* Signals sent by this widget */

enum {
       COMMIT,
       LAST_SIGNAL
     };

static gint gtk_programmateur_signals[LAST_SIGNAL] = { 0 };

/* Les donnes et fonctions statiques ici sont  comprendre comme des
   membres qui appartiennent  l'objet. Cependant, en ne faisant pas
   partie de la dfinition de la classe ou de la structure, elles
   deviennent invisibles de l'extrieures et, de l, PRIVEES ! */

static const bloc_cycle_t CYCLE_NUL = { { 0 }, { 0 } };
static GdkRectangle rectangle;
static GdkPixmap *the_pixmap;/* Pour le redessin de la the_widget */
static GdkCursor *resize_cursor, *notify_cursor;
static gint i, k, x, y,
       case_width, case_height, delta_x,
       left_limit, right_limit, new_debut, new_fin,
       pix_w, /* unit d'affichage des cycles (1/6) */
       fleche_clic[3];
static guint mouse_state, selected_cycle,
       in_center, in_left, in_right;
static div_t heure_calcul;
static gchar tampon[5];
static geometrie_t position_cycle[MAX_CYCLES];
static guint hh_win_width, hh_win_height, ok_btn_width, arrow_btn_width,
       haut_txt, larg_txt, clic_i;
static GdkGC *the_gc;
static GdkColor the_color;
static GdkColormap *the_colormap;

/*-------- Rfrences avant ----------------------------------*/

static GtkWidgetClass *parent_class = NULL;

static void gtk_programmateur_class_init(GtkProgrammateurClass *);
static void gtk_programmateur_init(GtkProgrammateur *);
static void gtk_programmateur_realize(GtkWidget *);
static void gtk_programmateur_unrealize(GtkWidget *);
static void gtk_programmateur_destroy( GtkObject * );

static void gtk_programmateur_size_request(GtkWidget *, GtkRequisition *);
static void gtk_programmateur_size_allocate(GtkWidget *, GtkAllocation *);

static void gtk_programmateur_paint(GtkWidget *, GdkRectangle *);
static void gtk_programmateur_draw(GtkWidget *, GdkRectangle *);
static gint gtk_programmateur_expose(GtkWidget *, GdkEventExpose *);

static gint gtk_programmateur_button_press(GtkWidget *, GdkEventButton *);
static gint gtk_programmateur_button_release(GtkWidget *, GdkEventButton *);
static gint gtk_programmateur_motion_notify(GtkWidget *, GdkEventMotion *);

/*------------- Fonctions de service : prive ----------------------------*/

inline static void map_horaires( GtkProgrammateur *programmateur,
                                 guint cycle_id )
{
  heure_calcul = div( 15 * position_cycle[cycle_id].x_debut, 60 );
  programmateur->cycle[cycle_id].heure_debut.heures = heure_calcul.quot;
  programmateur->cycle[cycle_id].heure_debut.minutes = heure_calcul.rem;

  heure_calcul = div( 15 * position_cycle[cycle_id].x_fin, 60 );
  programmateur->cycle[cycle_id].heure_fin.heures = heure_calcul.quot;
  programmateur->cycle[cycle_id].heure_fin.minutes = heure_calcul.rem;
}

/*--------------------------------------------------------------------------*/

inline static void add_new_cycle( GtkProgrammateur *programmateur )
{
  /* Achtung : la taille des cycles est exprime en 1/8 de case pour 15 min*/

  if ( programmateur->nb_cycles == 1 ) /* 1er cycle : ya la place */
     {
       position_cycle[0].x_debut = 0;
       position_cycle[0].x_fin = 8;
       position_cycle[0].x_width = 8;

       map_horaires( programmateur, 0 );
     }

  else if ( position_cycle[programmateur->nb_cycles - 2].x_fin
            < FULL_DAY - 8 ) /* Ya la place pour rajouter un cycle... */
     {
       position_cycle[programmateur->nb_cycles - 1].x_debut =
       position_cycle[programmateur->nb_cycles - 2].x_fin + 1;

       position_cycle[programmateur->nb_cycles - 1].x_width = 8;

       position_cycle[programmateur->nb_cycles - 1].x_fin =
       position_cycle[programmateur->nb_cycles - 1].x_debut +
       position_cycle[programmateur->nb_cycles - 1].x_width;

       map_horaires( programmateur, programmateur->nb_cycles - 1 );
     }
  else  /* No room for one more cycle : we just refuse the new one */
       {
         gdk_beep();
         programmateur->nb_cycles--;
       }
}

/*--------------------------------------------------------------------------*/

GtkType gtk_programmateur_get_type( void )
{
  static GtkType programmateur_type = 0;

  if ( !programmateur_type )
    {
      static const GtkTypeInfo programmateur_info =
      {
	      "GtkProgrammateur",
	      sizeof (GtkProgrammateur),
	      sizeof (GtkProgrammateurClass),
	      (GtkClassInitFunc)  gtk_programmateur_class_init,
	      (GtkObjectInitFunc) gtk_programmateur_init,
	      (GtkArgSetFunc) NULL,
	      (GtkArgGetFunc) NULL,
      };
      programmateur_type = gtk_type_unique( GTK_TYPE_WIDGET, &programmateur_info );
    }
  return programmateur_type;
}

/*-------------------------------------------------------------------------*/

static void gtk_programmateur_class_init( GtkProgrammateurClass *klass )
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;

  object_class = (GtkObjectClass*) klass;
  widget_class = (GtkWidgetClass*) klass;

  parent_class = gtk_type_class( GTK_TYPE_WIDGET );

  widget_class->realize   = gtk_programmateur_realize;
  widget_class->unrealize = gtk_programmateur_unrealize;

  object_class->destroy   = gtk_programmateur_destroy;

  widget_class->draw         = gtk_programmateur_draw;
  widget_class->expose_event = gtk_programmateur_expose;

  widget_class->size_request  = gtk_programmateur_size_request;
  widget_class->size_allocate = gtk_programmateur_size_allocate;

  widget_class->button_press_event   = gtk_programmateur_button_press;
  widget_class->button_release_event = gtk_programmateur_button_release;
  widget_class->motion_notify_event  = gtk_programmateur_motion_notify;

  /* Signal registering */

  gtk_programmateur_signals[COMMIT] =
    gtk_signal_new( "commit",
		                GTK_RUN_FIRST, object_class->type,
		                GTK_SIGNAL_OFFSET( GtkProgrammateurClass, commit ),
		                gtk_signal_default_marshaller, GTK_TYPE_NONE, 0 );

  gtk_object_class_add_signals( object_class,
                                gtk_programmateur_signals,
                                LAST_SIGNAL );

  /* klass->call backs = NULL pour annoncer qui ya pas
     de traitements par defaut  */

  klass->commit = NULL;
}

/*-------------------------------------------------------------------------*/

static void gtk_programmateur_init( GtkProgrammateur *programmateur )
{
  GtkWidget *the_widget;

  the_widget = GTK_WIDGET(programmateur);
  GTK_WIDGET_SET_FLAGS (the_widget, GTK_CAN_FOCUS);

  haut_txt = gdk_char_height( the_widget->style->font, '0' );
  larg_txt = gdk_string_width( the_widget->style->font, "Enter" );

  fleche_clic[0] = fleche_clic[1] = fleche_clic[2]= GTK_SHADOW_OUT;

  /* Pixmap image de la fentre. A l'init, il n'est pas encore rfrenc */

  the_pixmap = NULL;
  the_colormap = gdk_colormap_get_system(); /* For the change of colors */

  /* Initialisations relatives  la souris */

  resize_cursor = gdk_cursor_new( GDK_SB_H_DOUBLE_ARROW );
  notify_cursor = gdk_cursor_new( GDK_HAND2 );
  mouse_state = NO_STATE;
  in_left = in_center = in_right = FALSE;


  /* Initialisation des cycles et de leur position */

  for ( k = 0; k < MAX_CYCLES; k++ )
     {
       programmateur->cycle[k] = CYCLE_NUL;
       position_cycle[k].x_debut = position_cycle[k].x_width = 0;
     }
  programmateur->nb_cycles = 0;
}

/*-------------------------------------------------------------------------*/

static void gtk_programmateur_destroy( GtkObject *the_object )
{
  GtkProgrammateur *programmateur;

  g_return_if_fail( the_object != NULL );
  g_return_if_fail( GTK_IS_PROGRAMMATEUR(the_object) );

  programmateur = GTK_PROGRAMMATEUR(the_object);

  gdk_cursor_destroy(resize_cursor);
  gdk_cursor_destroy(notify_cursor);

  /* Chaining up for the parents */

  if ( GTK_OBJECT_CLASS(parent_class)->destroy )
    (* GTK_OBJECT_CLASS(parent_class)->destroy)(the_object);
}

/*-------------------------------------------------------------------------*/

GtkWidget * gtk_programmateur_new( void )
{
  GtkWidget *programmateur;
  programmateur = gtk_type_new( gtk_programmateur_get_type() );

  return GTK_WIDGET( programmateur );
}

/*-------------------------------------------------------------------------*/

static void gtk_programmateur_realize( GtkWidget *the_widget )
{
  GtkProgrammateur *programmateur;
  GdkWindowAttr attributes;
  gint attributes_mask;

  g_return_if_fail (the_widget != NULL);
  g_return_if_fail (GTK_IS_PROGRAMMATEUR (the_widget));

  GTK_WIDGET_SET_FLAGS (the_widget, GTK_REALIZED);

  programmateur = GTK_PROGRAMMATEUR (the_widget);

  attributes.x = the_widget->allocation.x;
  attributes.y = the_widget->allocation.y;
  attributes.width = the_widget->allocation.width;
  attributes.height = the_widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.event_mask = gtk_widget_get_events( the_widget ) | GDK_EXPOSURE_MASK |
                          GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
                          GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK;

  attributes.visual = gtk_widget_get_visual( the_widget );
  attributes.colormap = gtk_widget_get_colormap( the_widget );

  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
  the_widget->window = gdk_window_new( the_widget->parent->window, &attributes, attributes_mask );

  the_widget->style = gtk_style_attach( the_widget->style, the_widget->window );

  gdk_window_set_user_data( the_widget->window, the_widget );

  gtk_style_set_background( the_widget->style, the_widget->window, GTK_STATE_ACTIVE );
}

/*-------------------------------------------------------------------------*/

static void gtk_programmateur_unrealize (GtkWidget *the_widget)
{
  GtkProgrammateur *programmateur;

  g_return_if_fail( the_widget != NULL );
  g_return_if_fail( GTK_IS_PROGRAMMATEUR(the_widget) );

  programmateur = GTK_PROGRAMMATEUR(the_widget);

  if ( GTK_WIDGET_MAPPED(the_widget) )
     gtk_widget_unmap(the_widget);

  GTK_WIDGET_UNSET_FLAGS( the_widget, GTK_MAPPED );

  /* destroy the widget->window and unset the realized flag */

  if ( GTK_WIDGET_CLASS(parent_class)->unrealize )
    (* GTK_WIDGET_CLASS(parent_class)->unrealize)( the_widget );
}

/*-------------------------------------------------------------------------*/

static void gtk_programmateur_size_request(GtkWidget *the_widget, GtkRequisition *requisition)
{
  /* longueur de la petite fenetre qui affiche les horaires */

  hh_win_width = 2 + gdk_string_width( the_widget->style->font, "00:00" );
  hh_win_height = 2 + gdk_char_height( the_widget->style->font, '0' );

  /* On aligne la taille initiale  celle de la fenetre horaire texte en la
     faisant concider avec un multiple de 8 pour les cases de cycles. */

  pix_w = div( hh_win_width - 2, 8 ).quot;
  case_width = 8 * pix_w;
  requisition->width = 16 * case_width;
  requisition->height = 3 * gdk_char_height( the_widget->style->font, '0');

  /* A faire pour la premire allocation */

  arrow_btn_width = case_width;
  ok_btn_width = 2 * arrow_btn_width;
  case_height = the_widget->allocation.height;
}

/*-------------------------------------------------------------------------*/

static void gtk_programmateur_size_allocate(GtkWidget *the_widget, GtkAllocation *allocation)
{
  GtkProgrammateur *programmateur;
  div_t new_size;
  gint aa;

  g_return_if_fail (the_widget != NULL);
  g_return_if_fail (GTK_IS_PROGRAMMATEUR (the_widget));
  g_return_if_fail (allocation != NULL);

  the_widget->allocation = *allocation;

  if (GTK_WIDGET_REALIZED (the_widget))
    {
      programmateur = GTK_PROGRAMMATEUR (the_widget);

      new_size = div( allocation->width, 8 * 16 );

      if ( new_size.rem == 0 ) /* Ca tombe juste ? */
       {
         pix_w = new_size.quot;
         case_width = 8 * pix_w;

         arrow_btn_width = case_width;
         ok_btn_width = 2 * arrow_btn_width;
       }

      else /* Ca tombe pas juste : on equilibre. */
       {
         pix_w = new_size.quot;
         case_width = 8 * pix_w;

         aa = allocation->width - 12 * case_width;
         arrow_btn_width = div( aa, 4 ).quot;
         ok_btn_width = 2 * arrow_btn_width + div( aa, 4 ).rem;
       }

      case_height = allocation->height;

      gdk_window_move_resize( the_widget->window,
                              allocation->x, allocation->y,
                              allocation->width,
                              allocation->height );
    }
}

/*-------------------------------------------------------------------------*/

static gint gtk_programmateur_expose(GtkWidget *the_widget, GdkEventExpose *event)
{
  GtkProgrammateur *programmateur;

  g_return_val_if_fail (the_widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_PROGRAMMATEUR (the_widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  programmateur = GTK_PROGRAMMATEUR (the_widget);

  /* Recuperation de la taille de la fenetre */

  rectangle.x = 0;
  rectangle.y = 0;
  rectangle.width = the_widget->allocation.width;
  rectangle.height = the_widget->allocation.height;

  /* Appel de la fonction de dessin proprement dit */

  gtk_programmateur_paint( the_widget, &rectangle );

  return FALSE;
}

/*-------------------------------------------------------------------------*/

static void gtk_programmateur_draw( GtkWidget *the_widget, GdkRectangle *area )
{
  g_return_if_fail (the_widget != NULL);
  g_return_if_fail (GTK_IS_PROGRAMMATEUR (the_widget));
  g_return_if_fail (area != NULL);

  gtk_programmateur_paint( the_widget, area );
}

/*-------------------------------------------------------------------------*/

static void gtk_programmateur_paint( GtkWidget *the_widget, GdkRectangle *area )
{
  GtkProgrammateur *programmateur;
  GdkPoint fleche[3];

  g_return_if_fail (the_widget != NULL);
  g_return_if_fail (the_widget->window != NULL);
  g_return_if_fail (GTK_IS_PROGRAMMATEUR (the_widget));

  programmateur = GTK_PROGRAMMATEUR (the_widget);

  /* Redfinition d'un nouveau pixmap  la nouvelle taille de la widget */

  the_pixmap = gdk_pixmap_new( the_widget->window,
                               the_widget->allocation.width,
                               the_widget->allocation.height, -1 );
  /* we use a private gc based on the widget mid one for the forced colors */
  the_gc = gdk_gc_new( the_pixmap );
  gdk_gc_copy( the_gc, the_widget->style->mid_gc[the_widget->state] );

  /* Dessin du fond de la zone des cycles */

  /* Construction and allocation of the "cold" color */

  gdk_color_parse( "DeepSkyBlue4", &the_color );
  gdk_colormap_alloc_color( the_colormap, &the_color, FALSE, TRUE );

  gdk_gc_set_foreground( the_gc, &the_color );

  gdk_draw_rectangle( the_pixmap, the_gc, TRUE,
                      0, 0,
                      12 * case_width, case_height );

  gtk_draw_shadow( the_widget->style, the_pixmap, the_widget->state,
                   GTK_SHADOW_IN,
                      0, 0,
                      12 * case_width, case_height );

  /* freeing the cold color previously allocated */

  gdk_colormap_free_colors( the_colormap, &the_color, 1 );

  /* Dessin des cycles s'il y a lieu */

  if ( programmateur->nb_cycles != 0 )

    {
      /* Construction and allocation of the "warm" color */

      gdk_gc_copy( the_gc, the_widget->style->mid_gc[the_widget->state] );
      gdk_color_parse( "OrangeRed3", &the_color );
      gdk_colormap_alloc_color( the_colormap, &the_color, FALSE, TRUE );

      gdk_gc_set_foreground( the_gc, &the_color );

      for ( k = 0; k < programmateur->nb_cycles; k++ )
         {
           gdk_draw_rectangle( the_pixmap, the_gc, TRUE,
                               position_cycle[k].x_debut * pix_w,
                               2,
                               position_cycle[k].x_width * pix_w,
                               case_height - 3 );

           gtk_draw_shadow( the_widget->style, the_pixmap, the_widget->state,
                            GTK_SHADOW_OUT,
                            position_cycle[k].x_debut * pix_w,
                            2,
                            position_cycle[k].x_width * pix_w,
                            case_height - 3 );
         }

      /* freeing the "warm" color previously used for the cycle */

      gdk_colormap_free_colors( the_colormap, &the_color, 1 );
    }

  /* Dessin de flches de choix du nb de cycles */

  gdk_draw_rectangle( the_pixmap, the_widget->style->mid_gc[the_widget->state],
                      TRUE,
                      12 * case_width, 2,
                      arrow_btn_width, case_height - 2 );

  gtk_draw_shadow( the_widget->style, the_pixmap, the_widget->state,
                   fleche_clic[0], /* style shadow */
                   12 * case_width, 0,
                   arrow_btn_width, case_height );

  fleche[0].x = 12 * case_width + arrow_btn_width / 6;
  fleche[0].y = case_height / 2;
  fleche[1].x = 12 * case_width + 2 * arrow_btn_width / 3;
  fleche[1].y = case_height / 4;
  fleche[2].x = 12 * case_width + 2 * arrow_btn_width / 3;
  fleche[2].y = 3 * case_height / 4;

  gdk_draw_polygon( the_pixmap,
                    the_widget->style->fg_gc[the_widget->state],
                    TRUE, fleche, 3 );

  gdk_draw_rectangle( the_pixmap, the_widget->style->mid_gc[the_widget->state],
                      TRUE,
                      12 * case_width + arrow_btn_width, 2,
                      arrow_btn_width, case_height - 2 );

  gtk_draw_shadow( the_widget->style, the_pixmap, the_widget->state,
                   fleche_clic[1], /* style shadow */
                   12 * case_width + arrow_btn_width, 0,
                   arrow_btn_width, case_height );

  fleche[0].x = 12 * case_width + arrow_btn_width + arrow_btn_width / 3;
  fleche[0].y = case_height / 4;
  fleche[1].x = 12 * case_width + arrow_btn_width + 5 * arrow_btn_width / 6;
  fleche[1].y = case_height / 2;
  fleche[2].x = 12 * case_width + arrow_btn_width + arrow_btn_width / 3;
  fleche[2].y = 3 * case_height / 4;

  gdk_draw_polygon( the_pixmap,
                    the_widget->style->fg_gc[the_widget->state],
                    TRUE, fleche, 3 );

  /* Dessin du bouton Enter */

  gdk_draw_rectangle( the_pixmap, the_widget->style->mid_gc[the_widget->state],
                      TRUE,
                      12 * case_width + 2 * arrow_btn_width, 0,
                      ok_btn_width, case_height );

  gtk_draw_shadow( the_widget->style, the_pixmap, the_widget->state,
                   fleche_clic[2],
                   12 * case_width + 2 * arrow_btn_width, 0,
                   ok_btn_width, case_height );

  sprintf(tampon, "Enter" );

  gdk_draw_string( the_pixmap, the_widget->style->font,
                   the_widget->style->fg_gc[the_widget->state],
                   12 * case_width + 2 * arrow_btn_width +
                   ( ok_btn_width - larg_txt ) / 2,
                   haut_txt +
                   ( case_height - haut_txt ) / 2,
                   tampon );

  /* Dessin des notifications s'il y a lieu */

  if ( ( mouse_state >= LEFT_STATE ) && ( programmateur->nb_cycles > 0 ) )
   {
     /* Calcul des horaires relatifs  la gometrie */

     sprintf( tampon, "%.2d:%.2d",
              programmateur->cycle[selected_cycle].heure_debut.heures,
              programmateur->cycle[selected_cycle].heure_debut.minutes );

     gdk_draw_rectangle( the_pixmap, the_widget->style->white_gc, TRUE,
                         position_cycle[selected_cycle].x_debut * pix_w - hh_win_width / 2,
                         2,
                         hh_win_width, hh_win_height );

     gdk_draw_string( the_pixmap, the_widget->style->font,
                      the_widget->style->fg_gc[the_widget->state],
                      position_cycle[selected_cycle].x_debut * pix_w - hh_win_width / 2 + 1,
                      hh_win_height + 1,
                      tampon );

     sprintf( tampon, "%.2d:%.2d",
              programmateur->cycle[selected_cycle].heure_fin.heures,
              programmateur->cycle[selected_cycle].heure_fin.minutes );


     gdk_draw_rectangle( the_pixmap, the_widget->style->white_gc, TRUE,
                         position_cycle[selected_cycle].x_fin * pix_w - hh_win_width / 2,
                         case_height - hh_win_height - 2,
                         hh_win_width, hh_win_height );

     gdk_draw_string( the_pixmap, the_widget->style->font,
                      the_widget->style->fg_gc[the_widget->state],
                      position_cycle[selected_cycle].x_fin * pix_w - hh_win_width / 2 + 1,
                      case_height - 3,
                      tampon );
   }

  /* Blitting du pixmap  la fentre pour un affichage instantann. */

  gdk_draw_pixmap( the_widget->window,
                   the_widget->style->fg_gc[the_widget->state],
                   the_pixmap,
                   0, 0, 0, 0, -1, -1 );
  gdk_gc_unref(the_gc);
  gdk_pixmap_unref(the_pixmap);
}

/*-------------------------------------------------------------------------*/

static gint gtk_programmateur_button_press( GtkWidget *the_widget,
                                            GdkEventButton *event )
{
  GtkProgrammateur *programmateur;

  g_return_val_if_fail (the_widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_PROGRAMMATEUR (the_widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  programmateur = GTK_PROGRAMMATEUR(the_widget);

  clic_i = 4; /* Pas de release  prendre en compte */

  /* Identification de la case i clique */

  if ( mouse_state == NO_STATE )
   {
     i = div(event->x, case_width).quot;

     if ( i >= 12 ) /* It is a button */
      {
        i = event->x - 12 * case_width;
        if ( i <= arrow_btn_width ) /* Arrow -- button */
         {
           fleche_clic[0] = GTK_SHADOW_IN;
           clic_i = 0;
           gtk_programmateur_paint( the_widget, &rectangle );
           return FALSE;
         }
        else if ( i <= 2 * arrow_btn_width ) /* Arrow ++ button */
              {
                fleche_clic[1] = GTK_SHADOW_IN;
                clic_i = 1;
                gtk_programmateur_paint( the_widget, &rectangle );
                return FALSE;
              }
        else  /* Enter button */
              {
                fleche_clic[2] = GTK_SHADOW_IN;
                clic_i = 2;
                gtk_programmateur_paint( the_widget, &rectangle );
                return FALSE;
              }
      }
   }

  else  /* Actions pending... */
   {
     mouse_state = mouse_state | PRESS_STATE;
   }
  return FALSE;
}

/*-------------------------------------------------------------------------*/

static gint gtk_programmateur_button_release( GtkWidget *the_widget,
                                              GdkEventButton *event )
{
  GtkProgrammateur *programmateur;

  g_return_val_if_fail (the_widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_PROGRAMMATEUR (the_widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  programmateur = GTK_PROGRAMMATEUR(the_widget);

  /* Code modified in order to take the press effect at the release... */

   if ( clic_i < 3 ) /* release regards a button */
    {
      fleche_clic[clic_i] = GTK_SHADOW_OUT;

      switch( clic_i )
      {
        case 0:
           if ( programmateur->nb_cycles > 0 )
              programmateur->nb_cycles--;
        break;
        case 1:
           if ( programmateur->nb_cycles < MAX_CYCLES )
            {
              programmateur->nb_cycles++;
              add_new_cycle( programmateur );
            };
        break;
        default: /* i=2 and we face the release of the "ENTER" btn */
           gtk_signal_emit( GTK_OBJECT(the_widget),
                            gtk_programmateur_signals[COMMIT] );
        break;
      }
    }

   else if ( mouse_state >= PRESS_STATE ) /* End of action move-resize */
         {
           mouse_state = mouse_state & ~PRESS_STATE;
         }

   gtk_programmateur_paint( the_widget, &rectangle );
   return FALSE;
}

/*-------------------------------------------------------------------------*/

static gint gtk_programmateur_motion_notify(GtkWidget *the_widget,
                                            GdkEventMotion *event)
{
  GtkProgrammateur *programmateur;

  g_return_val_if_fail (the_widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_PROGRAMMATEUR (the_widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  programmateur = GTK_PROGRAMMATEUR(the_widget);

  /* Identification de la case o se trouve le pointeur */

  gdk_window_get_pointer(the_widget->window, &x, &y, NULL);

  /* Localisation du curseur dans le mme systme de coordonnes
     que les "botes" de cycle*/

  i = div(x, pix_w).quot; /* unit = case/6 = widget_width/15/6 */

  /* If no button pressed whilst the move the we just notify */

  if ( mouse_state < PRESS_STATE )
   {
     in_left = in_center = in_right = FALSE;
     selected_cycle = programmateur->nb_cycles;

     /* On recherche si le pointeur est sur un cycle ou dans son voisinage.
     S'il n'y a pas de cycle, la question se pose pas. Ne pas oublier l'chelle
     des valeurs qui s'exprime en 1/8 de la largeur d'une case...*/

     if ( selected_cycle > 0 )
      {
        do {
             selected_cycle--;

             in_left   = ( ( i <= position_cycle[selected_cycle].x_debut + 1 ) &&
                         ( i > position_cycle[selected_cycle].x_debut - 1 ) );
             in_center = ( ( i > position_cycle[selected_cycle].x_debut + 1 ) &&
                         ( i < position_cycle[selected_cycle].x_fin - 1 ) );
             in_right  = ( ( i < position_cycle[selected_cycle].x_fin + 1 ) &&
                         ( i >= position_cycle[selected_cycle].x_fin - 1 ) );

           }
        while ( !in_left && !in_center && !in_right && ( selected_cycle > 0 ) );

        if ( in_left )
         {
           mouse_state = LEFT_STATE;
           gdk_window_set_cursor( the_widget->window, resize_cursor );
           gtk_programmateur_paint( the_widget, &rectangle );
           return FALSE;
         }
        if ( in_center )
         {
           mouse_state = CENTER_STATE;
           gdk_window_set_cursor( the_widget->window, notify_cursor );
           gtk_programmateur_paint( the_widget, &rectangle );
           return FALSE;
         }
        if ( in_right )
         {
           mouse_state = RIGHT_STATE;
           gdk_window_set_cursor( the_widget->window, resize_cursor );
           gtk_programmateur_paint( the_widget, &rectangle );
           return FALSE;
         }
      }

  /* Le curseur n'est  aucun endroit relevant : on ne redessine
     pas la fentre aprs avoir rtablit le curseur par dfaut */

     mouse_state = NO_STATE;
     gdk_window_set_cursor( the_widget->window, NULL );
     return FALSE;
   }

  /* Ok button pressed : eigther we move or we resize */

  else
   {
     if ( mouse_state == ( LEFT_STATE | PRESS_STATE ) )
      {
        if (
             (
               ( ( selected_cycle == 0 ) && ( i >= 0 ) ) ||
               ( ( selected_cycle != 0 ) &&
                 ( i > position_cycle[selected_cycle-1].x_fin ) )
             ) && ( position_cycle[selected_cycle].x_fin - i >= 8 )
           )
          {
            position_cycle[selected_cycle].x_debut = i;
            position_cycle[selected_cycle].x_width =
                                   position_cycle[selected_cycle].x_fin - i;

            map_horaires( programmateur, selected_cycle );

            gtk_programmateur_paint( the_widget, &rectangle );
          }
        return FALSE;
      }
     if ( mouse_state == ( RIGHT_STATE | PRESS_STATE ) )
      {
        if (
             (
               ( ( selected_cycle == programmateur->nb_cycles - 1 ) &&
                 ( i <= FULL_DAY ) ) ||
               ( ( selected_cycle != programmateur->nb_cycles - 1 ) &&
                 ( i < position_cycle[selected_cycle+1].x_debut ) )
             ) && (  i - position_cycle[selected_cycle].x_debut >= 8 )
           )
          {
            position_cycle[selected_cycle].x_fin = i;
            position_cycle[selected_cycle].x_width =
                                 i - position_cycle[selected_cycle].x_debut;

            map_horaires( programmateur, selected_cycle );

            gtk_programmateur_paint( the_widget, &rectangle );
          }
        return FALSE;
      }

     /* Fine : we just move. But before performing anything, one must first
     determine wether the move is allowed or not regarding the limits of
     the neighbourgs or of the widget itself */

     delta_x = i - ( ( position_cycle[selected_cycle].x_fin +
                       position_cycle[selected_cycle].x_debut ) / 2 );

     if ( selected_cycle == 0 ) /* first left cycle */
       {
         left_limit = 0;

         if ( programmateur->nb_cycles == 1 )  /* case if the only one */
           right_limit = FULL_DAY;

         else
           right_limit = position_cycle[selected_cycle + 1].x_debut - 1;
       }
     else if ( selected_cycle == programmateur->nb_cycles - 1 ) /* last right cycle */
           {
             right_limit = FULL_DAY;
             left_limit = position_cycle[selected_cycle - 1].x_fin + 1;
           }
     else {
            left_limit = position_cycle[selected_cycle - 1].x_fin + 1;
            right_limit = position_cycle[selected_cycle + 1].x_debut - 1;
          }

     /* The move only occurs if the cycle stays between its limits */

     new_debut = position_cycle[selected_cycle].x_debut + delta_x;
     new_fin = position_cycle[selected_cycle].x_fin + delta_x;

     if ( ( new_debut >= left_limit ) && ( new_fin <= right_limit ) )
      {
         position_cycle[selected_cycle].x_debut = new_debut;
         position_cycle[selected_cycle].x_fin = new_fin;

         map_horaires( programmateur, selected_cycle );

         gtk_programmateur_paint( the_widget, &rectangle );
      }
     return FALSE;
   }
}

/*-------------------------------------------------------------------------*/

gint gtk_programmateur_remove_cycle( GtkProgrammateur *programmateur)
{
  g_return_val_if_fail( programmateur != NULL, -1 );
  g_return_val_if_fail( programmateur->nb_cycles > 0, -1 );

  programmateur->nb_cycles--;

  gtk_programmateur_paint( GTK_WIDGET(programmateur), &rectangle );

  return programmateur->nb_cycles;
}

/*-------------------------------------------------------------------------*/

gint gtk_programmateur_get_cycle( GtkProgrammateur *programmateur,
                                             guint  cycle_id,
                                      bloc_cycle_t *cycle_copy )
{
  g_return_val_if_fail( programmateur != NULL, -1 );
  g_return_val_if_fail( programmateur->nb_cycles > cycle_id, -1 );

  *cycle_copy = programmateur->cycle[cycle_id];

  return 0;
}

/*-------------------------------------------------------------------------*/

gint gtk_programmateur_get_nb_cycles( GtkProgrammateur *programmateur )
{
  g_return_val_if_fail( programmateur != NULL, -1 );

  return programmateur->nb_cycles;
}

/*-------------------------------------------------------------------------*/

gint gtk_programmateur_reset_cycles( GtkProgrammateur *programmateur )
{
  g_return_val_if_fail( programmateur != NULL, -1 );

  for ( k = 0; k < MAX_CYCLES; k++ )
     {
       programmateur->cycle[k] = CYCLE_NUL;
       position_cycle[k].x_debut = 0;
       position_cycle[k].x_fin = 0;
       position_cycle[k].x_width = 0;
     }

  programmateur->nb_cycles = 0;

  gtk_programmateur_paint( GTK_WIDGET(programmateur), &rectangle );

  return 0;
}

/*-------------------------------------------------------------------------*/

gint gtk_programmateur_add_cycle( GtkProgrammateur *programmateur,
                                      bloc_cycle_t  the_cycle )
{
  g_return_val_if_fail( programmateur != NULL, -1 );
  g_return_val_if_fail( programmateur->nb_cycles < MAX_CYCLES, -1 );

  /* position in case units. No test are yet performed if the cycle
    fits within the window or not. */

  position_cycle[programmateur->nb_cycles].x_debut =
                             the_cycle.heure_debut.heures * 4 +
                             div( the_cycle.heure_debut.minutes, 15 ).quot;

  position_cycle[programmateur->nb_cycles].x_fin =
                             the_cycle.heure_fin.heures * 4 +
                             div( the_cycle.heure_fin.minutes, 15 ).quot;

  position_cycle[programmateur->nb_cycles].x_width =
                          position_cycle[programmateur->nb_cycles].x_fin -
                          position_cycle[programmateur->nb_cycles].x_debut;

  programmateur->cycle[programmateur->nb_cycles] = the_cycle;

  programmateur->nb_cycles++;

  gtk_programmateur_paint( GTK_WIDGET(programmateur), &rectangle );

  return programmateur->nb_cycles;
}
