/***************************************************************************
                          maincallbacks.c  -  description
                             -------------------
    begin                : Mon Apr 16 2001
    copyright            : (C) 2001 by Dominique Laigle
    email                : lainal@gmx.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <maincallbacks.h>

/* Common variables useful for all the functions. */

static gint  i, j, k;
static guint nb_cycles, nb_copy, nb_paste, nb_select,
             day_id_src, day_id_dest,
             idx_day, idx_group, idx_cycle, idx_copy, idx_paste,
             days_select[31];
static const bloc_cycle_t  CYCLE_NUL = { { 0 }, { 0 } };
static GDate date_tampon,
             *dates_copy[31], *dates_paste[31];

/* Some usefull Functions */

/************************************************************************/

gint delete_event_cb( GtkWidget *appli_window, GdkEventAny *e, gpointer data)
{
  gtk_main_quit();
  return FALSE;
}

/*---------------------------------------------------------------------*/

void slot_import_btn_clicked( void )
{
  gint retour;
  gchar export_crontab[256];

  slot_reset_btn_clicked();

  /* building the unix system command to import the crontab */

  strcpy( export_crontab, "crontab -l | grep -v SHELL | grep -v MAILTO |" );
  strcat( export_crontab, " grep -v \'\\#\' | grep -v ^$ > " );
  strcat( export_crontab, CRON_IMPEX_CRONTAB_FILE );

  if ( system( export_crontab ) != 0 )
    {
       gdk_beep();
       fprintf( stderr, "Impossible to get the crontab\n" );
       return;
    }

  retour = cron_impex_import_crontab( CRON_IMPEX_CRONTAB_FILE, schedule );

  if ( retour != CRON_IMPEX_OK )
   {
     gdk_beep();
     fprintf( stderr, "Error in cron-impex : %s\n",
              cron_impex_verbose_error_message( retour ) );
   }

  /* Force the redrawing of the day to display the new program */

  gtk_calendrier_get_internal_date( my_calendrier, &date_tampon );
  idx_day = g_date_day_of_year( &date_tampon ) - 1;
  idx_group = my_groupeadmin->activ_group;

  nb_cycles = schedule[idx_day].day_program.nb_cycles[idx_group];

  gtk_programmateur_reset_cycles( my_programmateur );

  if ( nb_cycles != 0 )
     for ( idx_cycle = 0; idx_cycle < nb_cycles; idx_cycle++ )
        gtk_programmateur_add_cycle( my_programmateur,
           schedule[idx_day].day_program.day_cycles[idx_group][idx_cycle] );
}

/*---------------------------------------------------------------------*/

void slot_export_btn_clicked( void )
{
  gint retour;
  gchar insert_crontab[256];
  retour = cron_impex_export_crontab( schedule, CRON_IMPEX_CRONTAB_FILE );

  if ( retour != CRON_IMPEX_OK )
   {
     gdk_beep();
     cron_impex_verbose_error_message( retour );
   }
  else {
         strcpy( insert_crontab, "/usr/bin/crontab " );
         strcat( insert_crontab, CRON_IMPEX_CRONTAB_FILE );

         if ( system( insert_crontab ) != 0 )
          {
            gdk_beep();
            fprintf(stderr, "Impossible to insert the crontab\n");
          }
       }
}

/*---------------------------------------------------------------------*/

void slot_ejp_btn_clicked( void )
{
  /* This function was previously dedicated to the EJP mode. Since it is
     not usefull anymore, we replaced its purpose by an interface test.
   */

  gchar  env_valeur, *gp_var_env_nom, *gp_var_env_val,
         *gp_id = g_new0( gchar, 3 );
  gchar  *command = g_new0( gchar, 64 );
  guchar group_definition_mask[MAX_GROUPES], all_groups, status_if;
  gint   i, result, status_file;
  gboolean env_ok;
  GTimer *the_timer;

  /* Group definitions from the environment variables. If the environment
     is not set, then an error is raised and the application exits. */

  for ( idx_group = 0; idx_group < MAX_GROUPES; idx_group++ )
    {
       g_snprintf( gp_id, 2, "%d", idx_group );
       gp_var_env_nom = g_strconcat( "HEATHER_GP_", gp_id, NULL );

       gp_var_env_val = getenv( gp_var_env_nom );

       g_return_if_fail( gp_var_env_val != NULL );

       env_valeur = ( guchar ) strtol( gp_var_env_val, NULL, 0 );
       group_definition_mask[idx_group] = env_valeur;

       g_free( gp_var_env_nom );
    }

  g_free( gp_id );

  /* Test the coherence of the environment : all the bit assignments
     must be unique within the 4 groups. */

  all_groups = group_definition_mask[0];
  env_ok = TRUE;

  for ( i = 1; i < MAX_GROUPES; i++ )
    if ( ( all_groups & group_definition_mask[i] ) !=0 )
     {
       env_ok = FALSE;
       break;
     }
    else all_groups |= group_definition_mask[i];

  g_return_if_fail( env_ok == TRUE ); /* cycle definition is broken */

  /* get an fd for the status file opened RW */

  status_file = open( STATUS_FILE, O_RDWR | O_NDELAY,
                      S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
  if ( status_file < 0 )
    {
      fprintf( stderr, "\n%s : %s\n", STATUS_FILE, strerror( errno ) );
      return;
    }

  /* we read the actual interface status in order to put it back after
     the interface test. The test is not supposed to break the run of
     the eventually actual on going schedule ...
   */

  status_if = 0;
  result = read ( status_file, (void *) &status_if, 1 );
  if ( result < 0 )
    {
      close( status_file );
      fprintf( stderr, "\n%s : %s\n", STATUS_FILE, strerror( errno ) );
      return;
    }

  close( status_file );

  /* Let's play the group definition */

  the_timer = g_timer_new();
  g_timer_start( the_timer );

  for ( i = 0; i < MAX_GROUPES; i++ )
    {
      g_timer_reset( the_timer ); /* Every loop dures somewhat 1 second */

      sprintf( command, "/usr/bin/heathout %d", group_definition_mask[i] );
      system( command );
      slot_status_btn_clicked();

      /* Since we'll lock the execution whilst 1 second, flush all pending
         events ( such as widgets redrawing ) to be sure that GTK has
         nothing to do anymore. */

      while ( gtk_events_pending() )
           gtk_main_iteration();

      /* Let's wait up to one second before the next group... */

      while ( g_timer_elapsed( the_timer, NULL ) < 1.0 );
    }

  sprintf( command, "/usr/bin/heathout %d", status_if );
  system( command );

  g_timer_stop( the_timer );
  g_timer_destroy( the_timer );

  g_free( command );

  slot_status_btn_clicked(); /* to show that everything is back... */
}

/*---------------------------------------------------------------------*/

void slot_reset_btn_clicked( void )
{
  /* All the dates are cleared and then set to the 1st january
    of the current year */

  gtk_calendrier_get_internal_date( my_calendrier, &date_tampon );
  g_date_set_month( &date_tampon, 1);
  g_date_set_day( &date_tampon, 1);

  for (  i = 0; i < 366; i++ )
    {
      g_date_clear( &schedule[i].date, 1 );
      g_date_set_julian( &schedule[i].date, g_date_julian( &date_tampon ) );
      for ( j = 0; j < MAX_GROUPES; j++ )
        {
          schedule[i].day_program.nb_cycles[j] = 0;

          for ( k = 0; k < MAX_CYCLES; k++ )
            schedule[i].day_program.day_cycles[j][k] = CYCLE_NUL;
        }
    }

  gtk_programmateur_reset_cycles( my_programmateur );

  slot_status_btn_clicked();
}

/*---------------------------------------------------------------------*/

void slot_status_btn_clicked( void )
{
  guchar   group_definition_mask[MAX_GROUPES], all_groups, status_if;
  gchar    env_valeur, *gp_var_env_nom, *gp_var_env_val,
           *gp_id = g_new0( gchar, 3 );
  gint     status_file, gp_state, result;
  gboolean env_ok;

  /* Retrieving the group definition in the environement */

  for ( i = 0; i < MAX_GROUPES; i++ )
    {
       g_snprintf( gp_id, 2, "%d", i );
       gp_var_env_nom = g_strconcat( "HEATHER_GP_", gp_id, NULL );

       gp_var_env_val = getenv( gp_var_env_nom );
       g_return_if_fail( gp_var_env_val != NULL );

       env_valeur = ( guchar ) strtol( gp_var_env_val, NULL, 0 );
       group_definition_mask[i] = env_valeur;

       g_free( gp_var_env_nom );
    }
  g_free( gp_id );

  all_groups = group_definition_mask[0];
  env_ok = TRUE;

  for ( i = 1; i < MAX_GROUPES; i++ )
    if ( ( all_groups & group_definition_mask[i] ) !=0 )
     {
       env_ok = FALSE;
       break;
     }
    else all_groups |= group_definition_mask[i];

  g_return_if_fail( env_ok == TRUE ); /* cycle definition is broken */

  /* Reading the status file */

  status_file = open( STATUS_FILE, O_RDONLY | O_NDELAY );
  g_return_if_fail( status_file >= 0 );

  status_if = 0;
  result = read ( status_file, (void *) &status_if, 1 );
  close( status_file );
  g_return_if_fail( result >= 0 );

  /* display on flag binaire */

  for ( i = 0; i < MAX_GROUPES; i++ )
    {
      gp_state = status_if & group_definition_mask[i];

      if ( gp_state )
        gtk_flag_binaire_set_flag( my_flagbinaire, i );
      else
        gtk_flag_binaire_clear_flag( my_flagbinaire, i );
    }
}

/*---------------------------------------------------------------------*/

void slot_calendrier_new_date_selected( void )
{
  gtk_calendrier_get_internal_date( my_calendrier, &date_tampon );
  idx_day = g_date_day_of_year( &date_tampon ) - 1;

  idx_group = my_groupeadmin->activ_group;

  nb_cycles = schedule[idx_day].day_program.nb_cycles[idx_group];

  gtk_programmateur_reset_cycles( my_programmateur );

  if ( nb_cycles != 0 )
     for ( idx_cycle = 0; idx_cycle < nb_cycles; idx_cycle++ )
        gtk_programmateur_add_cycle( my_programmateur,
           schedule[idx_day].day_program.day_cycles[idx_group][idx_cycle] );
}

/*-------------------------------------------------------------------------*/

void slot_programmateur_commit( void )
{
  gtk_calendrier_get_internal_date( my_calendrier, &date_tampon );

  idx_group = my_groupeadmin->activ_group; /* GroupeAdmin has no IF... */

  nb_cycles = gtk_programmateur_get_nb_cycles( my_programmateur );

  if ( nb_cycles == 0 ) /* nothing to do ? */
    return; /* ...goodbye. */

  /* Now, we have to see to who the commit apply (new functionality):
    - If there is a multi-selection activated in the gtkclalendrier
      then the commit occurs in all days belonging to that selection.
    - If not, the commit only occurs in the current day.
  */

  nb_select = gtk_calendrier_get_days_selected( my_calendrier, days_select );

  if ( nb_select == 0 ) /* No selection : current day */
    {
      idx_day = g_date_day_of_year( &date_tampon ) - 1;

      g_date_set_julian( &schedule[idx_day].date,
                         g_date_julian( &date_tampon ) );

      schedule[idx_day].day_program.nb_cycles[idx_group] = nb_cycles;

      for ( idx_cycle = 0; idx_cycle < nb_cycles; idx_cycle++ )
         gtk_programmateur_get_cycle( my_programmateur, idx_cycle,
           &schedule[idx_day].day_program.day_cycles[idx_group][idx_cycle] );
    }
  else /* all the days become that program to commit */
    for ( k = 0; k < nb_select; k++ )
      {
        g_date_set_day( &date_tampon, days_select[k] );
        idx_day = g_date_day_of_year( &date_tampon ) - 1;

        g_date_set_julian( &schedule[idx_day].date,
                           g_date_julian( &date_tampon ) );

        schedule[idx_day].day_program.nb_cycles[idx_group] = nb_cycles;

        for ( idx_cycle = 0; idx_cycle < nb_cycles; idx_cycle++ )
          gtk_programmateur_get_cycle( my_programmateur, idx_cycle,
               &schedule[idx_day].day_program.day_cycles[idx_group][idx_cycle] );
      }
}

/*---------------------------------------------------------------------*/

void slot_new_group_defined( void )
{
  /* The new group definitions are transfered to the environment.
     All other routines that need these definitions refer only to
     the environment. If not defined, they yell error.
  */

  gchar *gp_var_env_nom,
        *gp_var_env_val,
        *gp_id;

  gp_id = g_new0( gchar, 3 );
  gp_var_env_val = g_new0( gchar, 8 );

  for ( i = 0; i < MAX_GROUPES; i++ )
    {
       g_snprintf( gp_id, 2, "%d", i );
       gp_var_env_nom = g_strconcat( "HEATHER_GP_", gp_id, NULL );

       g_snprintf( gp_var_env_val, 7, "%#.2x",
                   my_groupeadmin->group_definition_mask[i] );

       if ( setenv( gp_var_env_nom, gp_var_env_val, TRUE ) < 0 )
        {
          fprintf( stderr, "prb with environment\n" );

          g_free( gp_var_env_nom );
          g_free( gp_var_env_val );
          g_free( gp_id );

          return;
        }
       g_free( gp_var_env_nom );
    }

  g_free( gp_var_env_val );
  g_free( gp_id );
}

/*-------------------------------------------------------------------------*/

void slot_calendrier_new_date_paste( void )
{
  /*
   The clipboard policy is to copy and paste the FOUR GROUPS together.
   In other words, there is no way to copy from one group to an other.
   This goes toward the direction of the functional design :
   the groups were specified to permit different programs within a day.
  */

  nb_copy  = gtk_calendrier_get_clipboard_copy( my_calendrier, dates_copy );
  nb_paste = gtk_calendrier_get_clipboard_paste( my_calendrier, dates_paste );

  idx_copy = idx_paste = 0;

  while ( ( idx_copy < nb_copy ) && ( idx_paste < nb_paste ) )
  {
    /* what do we have in the clipboard ? */

    g_return_if_fail( g_date_valid( dates_copy[idx_copy] ) );
    g_return_if_fail( g_date_valid( dates_paste[idx_paste] ) );

    /* The dates become id's */

    day_id_src = g_date_day_of_year( dates_copy[idx_copy] ) - 1;
    day_id_dest = g_date_day_of_year( dates_paste[idx_paste] ) - 1;

    schedule[day_id_dest] = schedule[day_id_src];

    /* The date is not the one copied but the one related to the id */

    g_date_set_julian( &schedule[day_id_dest].date,
                       g_date_julian( dates_paste[idx_paste] ) );

    idx_copy++;
    idx_paste++;
  }

  /* So far we just copied as much as possible. Now, we have to see
     the cases where nb_copy and nb_paste are different:
     - nb_copy > nb_paste : obviously nothing to do...
     - nb_copy < nb_paste : we make multiple copy of the last one.
   */

  if ( nb_copy < nb_paste )
   {
     day_id_src--;
     for ( ; idx_paste < nb_paste; idx_paste++ )
       {
         g_return_if_fail( g_date_valid( dates_paste[idx_paste] ) );

         day_id_dest = g_date_day_of_year( dates_paste[idx_paste] ) - 1;

         schedule[day_id_dest] = schedule[day_id_src];

         g_date_set_julian( &schedule[day_id_dest].date,
                            g_date_julian( dates_paste[idx_paste] ) );
       }
   }

  /* The last copied day is set as selected */

  idx_group = my_groupeadmin->activ_group;
  idx_day = day_id_dest;

  gtk_calendrier_set_internal_date( my_calendrier, &schedule[idx_day].date );

  nb_cycles = schedule[idx_day].day_program.nb_cycles[idx_group];

  gtk_programmateur_reset_cycles( my_programmateur );

  if ( nb_cycles != 0 )
     for ( idx_cycle = 0; idx_cycle < nb_cycles; idx_cycle++ )
        gtk_programmateur_add_cycle( my_programmateur,
           schedule[idx_day].day_program.day_cycles[idx_group][idx_cycle] );
}
