/***************************************************************************
                          main.c  -  description
                             -------------------
    begin                : ven nov 12 09:37:51 CET 1999
    copyright            : (C) 1999 by Dominique Laigle
    email                : dom@seriat.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/***************************************************************************/
/*                                                                         */
/*    INDIQUE EN CLAIR L'ETAT DE L'INTERFACE SELON LE FICHIET DE STATUT    */
/*                                                                         */
/***************************************************************************/

/**
  *@author Dominique Laigle
  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define STATUS_FILE  "/tmp/gheather.status"
#define MAX_GROUPES  4

#include <glib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

/* Variables globales */

gint fichier,    /* descripteur du fichier d'tat de l'interface       */
     interface,  /* descripteur du fichier spcial du device parallle */
     resultat;   /* rsultat des appels des fonctions systme          */

/* Variables globales */

gchar    env_valeur, *gp_var_env_nom, *gp_var_env_val,*etat, *gp_id;
guchar   all_groups, statut_if,
         idx_group,groupe[MAX_GROUPES],
         group_definition_mask[MAX_GROUPES] = { 0 };
gboolean env_ok;

/*--------------------------------------------------------------------*/
/* Fonction principale. Vu la taille du programme, c'est la seule.    */
/*--------------------------------------------------------------------*/

int main(int argc, char ** argv)
{
  /* Commenons par vrifier comment on a t appel.*/

  if ( argc > 2 ) /* le 1er argument est le nom du programme lui-mme..*/
    {
      printf( "\nUsage : %s [<status file>]\n\n", argv[0] );
      return -1;
    }

  /* Rcupration del'environement */

  gp_id = g_new0( gchar, 3 );

  for ( idx_group = 0; idx_group < MAX_GROUPES; idx_group++ )
    {
       g_snprintf( gp_id, 2, "%d", idx_group );
       gp_var_env_nom = g_strconcat( "HEATHER_GP_", gp_id, NULL );

       gp_var_env_val = getenv( gp_var_env_nom );

       if ( gp_var_env_val == NULL )
        {
           fprintf( stderr,
                    "Missing environment HEATHER_GP_%d\n", idx_group );
           return -1;
        }

       env_valeur = ( guchar ) strtol( gp_var_env_val, NULL, 0 );
       group_definition_mask[idx_group] = env_valeur;

       g_free( gp_var_env_nom );
    }

  g_free( gp_id );

  /* Test the coherence of the environment : all the bit assignments
     must be unique within the 4 groups. */

  all_groups = group_definition_mask[0];
  env_ok = TRUE;

  for ( idx_group = 1; idx_group < MAX_GROUPES; idx_group++ )
    if ( ( all_groups & group_definition_mask[idx_group] ) !=0 )
     {
       env_ok = FALSE;
       break;
     }
    else all_groups |= group_definition_mask[idx_group];

  if(  env_ok != TRUE ) /* cycle definition is broken */
   {
     fprintf( stderr, "Group definitions corrupted\n" );
     return -1;
   }


  /* Ouverture R du fichier d'tat de l'interface. Si aucun nom de fichier de
  donn en argument, alors on prend le fichier dfinit par dfaut.*/

  if ( argc == 1 )
    fichier = open( STATUS_FILE, O_RDONLY | O_NDELAY );
   
  else fichier = open( argv[1], O_RDONLY | O_NDELAY );

  if ( fichier < 0 )
    {
      fprintf( stderr, "\n%s : %s\n", argv[1], strerror(errno) );
      return -1;
    }

  /* Lecture de l'tat actuel de l'interface depuis le fichier de statut. */

  statut_if = 0;
  resultat = read ( fichier, (void *) &statut_if, 1 );

  if ( resultat < 0 )
    {
      close(fichier);
      fprintf( stderr, "\n%s : %s\n", argv[1], strerror(errno) );
      return -1;
    }

  for ( idx_group= 0; idx_group < MAX_GROUPES; idx_group++ )
    groupe[idx_group] = statut_if & group_definition_mask[idx_group];


  fprintf( stdout, "\n--------------------------------------------------");
  fprintf( stdout, "\nEtat des radiateurs d'aprs le fichier de statut :");
  fprintf( stdout, "\n--------------------------------------------------\n");

  for ( idx_group = 0; idx_group < 4; idx_group++)
      { 
        etat = ( groupe[idx_group] )? "allums  " : "en veille";
        fprintf( stdout, "\n  - Les radiateurs du groupe %d sont %s ;",
                 idx_group + 1, etat);
      }

  fprintf( stdout, "\n\n--------------------------------------------------\n\n");


  /* Fini : fermeture des fichiers et au revoir ... */

  close(fichier);
  return 0;
}