/***************************************************************************
                          main.c  -  description
                             -------------------
    begin                : dim jui  8 21:16:34 CEST 2001
    copyright            : (C) 2001 by D.Laigle
    email                : lainal@gmx.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define STATUS_FILE  "/tmp/gheather.status"
#define INTERFACE    "/dev/lp0"

#define MAX_GROUPES  4

#include <glib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

/* Variables globales */

gint fichier,    /* descripteur du fichier d'tat de l'interface       */
     interface,  /* descripteur du fichier spcial du device parallle */
     resultat;   /* rsultat des appels des fonctions systme          */

/* Variables globales */

gchar    env_valeur, *gp_var_env_nom, *gp_var_env_val,*etat, *gp_id;
guchar   all_groups, statut_if, commande, action_on,
         idx_group,groupe[MAX_GROUPES],
         group_definition_mask[MAX_GROUPES] = { 0 };
gboolean env_ok;

/*--------------------------------------------------------------------*/

static int get_environment()
{
  /* Rcupration del'environement */

  gp_id = g_new0( gchar, 3 );

  for ( idx_group = 0; idx_group < MAX_GROUPES; idx_group++ )
    {
       g_snprintf( gp_id, 2, "%d", idx_group );
       gp_var_env_nom = g_strconcat( "HEATHER_GP_", gp_id, NULL );

       gp_var_env_val = getenv( gp_var_env_nom );

       if ( gp_var_env_val == NULL )
        {
           fprintf( stderr,
                    "Missing environment HEATHER_GP_%d\n", idx_group );
           return -1;
        }

       env_valeur = ( guchar ) strtol( gp_var_env_val, NULL, 0 );
       group_definition_mask[idx_group] = env_valeur;

       g_free( gp_var_env_nom );
    }

  g_free( gp_id );

  /* Test the coherence of the environment : all the bit assignments
     must be unique within the 4 groups. */

  all_groups = group_definition_mask[0];
  env_ok = TRUE;

  for ( idx_group = 1; idx_group < MAX_GROUPES; idx_group++ )
    if ( ( all_groups & group_definition_mask[idx_group] ) !=0 )
     {
       env_ok = FALSE;
       break;
     }
    else all_groups |= group_definition_mask[idx_group];

  if(  env_ok != TRUE ) /* cycle definition is broken */
   {
     fprintf( stderr, "Group definitions corrupted\n" );
     return -1;
   }
  return 0;
}

/*--------------------------------------------------------------------*/

int main(int argc, char ** argv)
{
  /* How were we called ? */

  if (argc < 2) /* le 1er argument est le nom du programme lui-mme..*/
    {
      fprintf( stderr, "\nUsage : %s <id_groupe> < on || off >", argv[0] );
      fprintf( stderr, "\n        %s <encoded action>\n\n", argv[0] );
      return -1;
    }

  commande = atoi( argv[1] );

  /* Rcupration et contrle de la validit du premier argument */

  if ( argc == 3 ) /* Appel du premier type */
    {

      if ( get_environment() != 0 )
        return -1;

      idx_group = atoi( argv[1] );
      if ( ( idx_group < 1 ) || ( idx_group > 4 ) )
        {
          fprintf( stderr, "\n id_groupe doit tre compris entre 1 et 4\n");
          return -1;
        }

      /* Rcupration et contrle de la validit du deuxime argument */

      if ( !strcmp( argv[2],"on" ) )
      action_on = 1;
      else if ( !strcmp( argv[2], "off") )
             action_on = 0;
           else {
                  fprintf( stderr, "\nL'action est soit on, soit off\n\n");
                  return -1;
                }

      /* Ouverture-cration R/W (R pour other) du fichier d'tat de l'interface. */

      fichier = open( STATUS_FILE,
                  O_RDWR | O_CREAT | O_NDELAY,
                  S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
      if ( fichier < 0 ) /* Y manquerait pas un setuid root... ? */
        {
          fprintf( stderr, "\n%s : %s\n", STATUS_FILE, strerror(errno) );
          return -1;
        }

      /* Ouverture du fichier spcial de l'interface. */

      interface = open( INTERFACE, O_WRONLY | O_NDELAY );
      if ( interface < 0 ) /* L non-plus... ? */
        {
          close(fichier);
          fprintf( stderr, "\n%s : %s\n", INTERFACE, strerror(errno) );
          return -1;
        }

      /* Lecture de l'tat actuel de l'interface depuis le fichier de statut. */

      statut_if = 0;
      resultat = read ( fichier, (void *) &statut_if, 1 );
      if ( resultat < 0 )
        {
          close(fichier);
          close(interface);
          fprintf( stderr, "\n%s : %s\n", STATUS_FILE, strerror(errno) );
          return -1;
        }

      /* Replacement de la "tte" au dbut du fichier suite  la lecture,
      de faon  ce que l'criture remplace la donne prcdente.     */

      resultat = lseek( fichier, 0, SEEK_SET);
      if ( resultat < 0 )
        {
          close(fichier);
          close(interface);
          fprintf( stderr, "\n%s : %s\n", STATUS_FILE, strerror(errno) );
          return -1;
        }

      /* Dtermination du code  envoyer vers l'interface */

      commande = group_definition_mask[idx_group - 1];

      if ( action_on == 1 )
          commande = statut_if | commande;  /* les "1" sont forcs */
      else
          commande = statut_if & ~commande; /* les "0" sont forcs */
    }

  else // C'est une commande directe que l'on effectue sans contrle (mode automatique)
    {
      commande = atoi( argv[1] );

      /* Ouverture-cration R/W (R pour other) du fichier d'tat de l'interface. */

      fichier = open( STATUS_FILE,
                  O_RDWR | O_CREAT | O_NDELAY,
                  S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
      if ( fichier < 0 ) /* Y manquerait pas un setuid root... ? */
        {
          fprintf( stderr, "\n%s : %s\n", STATUS_FILE, strerror(errno) );
          return -1;
        }

      /* Ouverture du fichier spcial de l'interface. */

      interface = open( INTERFACE, O_WRONLY | O_NDELAY );
      if ( interface < 0 ) /* L non-plus... ? */
        {
          close(fichier);
          fprintf( stderr, "\n%s : %s\n", INTERFACE, strerror(errno) );
          return -1;
        }
    }

  /* Portion de code commune aux deux types d'appel : on envoie l'octet vers l'interface
     et vers le fichier d'tat. */

  /* Envoi de la commande vers l'interface */

  resultat = write( interface, &commande, 1 );
  if ( resultat < 0 ) /* Pourquoi on peut pas crire vers le fichier spcial de l'I/F ? */
    {
      close(fichier);
      close(interface);
      fprintf( stderr, "\n%s : %s\n", INTERFACE, strerror(errno) );
      return -1;
    }

  /* Mise  jour du fichier d'tat */

  resultat = write( fichier, &commande, 1 );
  if ( resultat < 0 ) /* Pourquoi on peut pas crire vers le fichier de status ? */
    {
      close(fichier);
      close(interface);
      fprintf( stderr, "\n%s : %s\n", STATUS_FILE, strerror(errno) );
      return -1;
    }

  /* Fini : fermeture des fichiers et au revoir ... */

  close(fichier);
  close(interface);
  return 0;
}
